/*****************************************************************************
 * map.cpp - map class, the functions to handle a already loaded map goes here
 *
 ****************************************************************************/
#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif

#include <iostream>
#include "SDL.h"
#include "freecnc.h"
#include "map.h"
#include "unit.h"
#include "structure.h"
#include "path.h"
#include "config.h"
#include "loadingscreen.h"
#include "logger.h"
#include "netconnection.h"
#include "memtrack.h"

using namespace std;


/** Get the imagepool, called by the imagecache.
 * @returns the imagepool.
 */

vector<SHPImage *> *CnCMap::getImagePool()
{
    return &imagepool;
}


Uint32 CnCMap::getOverlay(Uint32 pos)
{
    Uint32 overlay;

    if( overlaymatrix[pos] & HAS_OVERLAY ) {
        overlay = overlays[pos]<<16;
    } else
        overlay = 0;
    return overlay;
}

Uint32 CnCMap::getTerrain(Uint32 pos, Sint16 &xoff, Sint16 &yoff)
{
    Uint32 terrain;

    if( overlaymatrix[pos] & HAS_TERRAIN ) {
        terrain = terrains[pos].shpnum;
        xoff = terrains[pos].xoffset;
        yoff = terrains[pos].yoffset;
    } else
        terrain = 0;
    return terrain;
}



/** Constructor that loads a map cmoplete with terrain, overlays, units
 * and structures.
 * @param the eventqueue.
 * @param the arguments object
 * @param the socket for network play
 */
CnCMap::CnCMap(ActionEventQueue *evQueue, NetConnection *connection, LoadingScreen *lscreen)
{
    Uint16 i, len, result;
    ConfigType config;
    char* buff;
    /* Copy the name of the map so we can store it */
    config = getConfig();
    strncpy( this->mapname, config.map, 9 );
    this->evQueue = evQueue;
    this->maptype = config.gamenum;
    this->gamemode = config.gamemode;
    this->connection = connection;
    scrollstep = config.scrollstep;
    scrolltime = config.scrolltime;
    maxscroll    = config.maxscroll;
    timeline = NULL;
    /* start at top right corner of map. */
    // the startpos for the map is stored in position 0
    scrollpos.curx = 0;
    scrollpos.cury = 0;
    scrollpos.curxtileoffs = 0;
    scrollpos.curytileoffs = 0;
    // FIXME: work out which directions are valid once [xy]pos are properly set.
    valscroll = SCROLL_RIGHT|SCROLL_DOWN;
        
    scrollvec.x = 0;
    scrollvec.y = 0;
    scrollvec.t = 0;
    toscroll    = false;
    for (i=0;i<NUMMARKS;++i) {
        scrollbookmarks[i].x = 0;
        scrollbookmarks[i].y = 0;
        scrollbookmarks[i].xtile = 0;
        scrollbookmarks[i].ytile = 0;
    }
    /* Load the ini part of the map */
    buff = new char[128];
    sprintf(buff,"Reading %s.INI",mapname);
    lscreen->setCurrentTask(buff);
    loadIni();

    /* Load the bin part of the map ( the tiles ) */
    sprintf(buff,"Loading %s.BIN",mapname);
    lscreen->setCurrentTask(buff);
    if (maptype == GAME_TD)
        loadBin();
    delete[] buff;

    //   Path::setMapSize(width, height);
    ppool->setAlliances();
    minimap = NULL;
    timeline = new TimeLine(ppool);
    timeline->update(ppool);
    if (gamemode == 2) {
        buff = new char[8];
        strncpy(buff,"ready\n",8);
        len = strlen(buff);
        result = connection->send((Uint8 *)buff, len);
        if (result < len) {
            logger->error("error occurred when sending to server.\n");
            throw new LoadMapError();
        }
        delete[] buff;
    }
}



/** Destructor, free up some memory */
CnCMap::~CnCMap()
{
    unsigned int i;

    delete[] brief;
    delete[] action;
    delete[] player;
    delete[] theater;
    delete[] theme;
    delete[] winmov;
    delete[] losemov;
    delete timeline;
    delete[] tilematrix;
    delete[] terraintypes;

    for( i = 0; i < tileimages.size(); i++ )
        SDL_FreeSurface(tileimages[i]);

    delete[] overlaymatrix;

    for( i = 0; i < imagepool.size(); i++ )
        delete imagepool[i];

    delete uspool;
    delete ppool;
    if (minimap != NULL) {
        SDL_FreeSurface(minimap);
    }
    for( i = 0; i < numShadowImg; i++ ){
        SDL_FreeSurface(shadowimages[i]);
    }
    delete[] shadowimages;
}



/** sets the scroll to the specified direction.
 * @param direction to scroll in.
 */

bool CnCMap::accScroll(Uint8 direction)
{
    Sint8 valids = 0;
    if (direction & SCROLL_UP) {
        if (scrollvec.y >= 0)
            scrollvec.y = -scrollstep;
        else if (scrollvec.y > -maxscroll)
            scrollvec.y -= scrollstep;
        if (valscroll & SCROLL_UP) {
            ++valids;
        } else {
            scrollvec.y = 0;
            --valids;
        }
    }
    if (direction & SCROLL_DOWN) {
        if (scrollvec.y <= 0)
            scrollvec.y = scrollstep;
        else if (scrollvec.y < maxscroll)
            scrollvec.y += scrollstep;
        if (valscroll & SCROLL_DOWN) {
            ++valids;
        } else {
            scrollvec.y = 0;
            --valids;
        }
    }
    if (direction & SCROLL_LEFT) {
        if (scrollvec.x >= 0)
            scrollvec.x =  -scrollstep;
        else if (scrollvec.x > -maxscroll)
            scrollvec.x -= scrollstep;
        if (valscroll & SCROLL_LEFT) {
            ++valids;
        } else {
            scrollvec.x = 0;
            --valids;
        }
    }
    if (direction & SCROLL_RIGHT) {
        if (scrollvec.x <= 0)
            scrollvec.x = scrollstep;
        else if (scrollvec.x < maxscroll)
            scrollvec.x += scrollstep;
        if (valscroll & SCROLL_RIGHT) {
            ++valids;
        } else {
            scrollvec.x = 0;
            --valids;
        }
    }
    if (valids < 0) {
        return toscroll = false;
    }
    scrollvec.t = 0;
    return toscroll = true;
}

bool CnCMap::absScroll(Sint16 dx, Sint16 dy)
{
    Sint8 valid = 0;
    const float fmax = (float)maxscroll/100.0f;
    if (dx < 0) {
        if (valscroll & SCROLL_LEFT) {
            ++valid;
            scrollvec.x = (Sint8)(MIN(dx,100) * fmax);
        } else {
            --valid;
            scrollvec.x = 0;
        }
    } else if (dx > 0) {
        if (valscroll & SCROLL_RIGHT) {
            ++valid;
            scrollvec.x = (Sint8)(MIN(dx,100) * fmax);
        } else {
            --valid;
            scrollvec.x = 0;
        }
    }
    if (dy < 0) {
        if (valscroll & SCROLL_UP) {
            ++valid;
            scrollvec.y = (Sint8)(MIN(dy,100) * fmax);
        } else {
            --valid;
            scrollvec.y = 0;
        }
    } else if (dy > 0) {
        if (valscroll & SCROLL_DOWN) {
            ++valid;
            scrollvec.y = (Sint8)(MIN(dy,100) * fmax);
        } else {
            --valid;
            scrollvec.y = 0;
        }
    }
    if (valid < 0) {
        return toscroll = false;
    }
    return toscroll = true;
}


/** scrolls according to the scroll vector.
*/
void CnCMap::doscroll()
{
    Sint32 xtile, ytile;
    if( scrollpos.curx*scrollpos.tilewidth+scrollpos.curxtileoffs <= -scrollvec.x && 
            scrollvec.x < 0){
        scrollvec.t = 0;
        scrollvec.x = 0;
        scrollpos.curx = 0;
        scrollpos.curxtileoffs = 0;
    }
    if( scrollpos.cury*scrollpos.tilewidth+scrollpos.curytileoffs <= -scrollvec.y && 
            scrollvec.y < 0){
        scrollvec.t = 0;
        scrollvec.y = 0;
        scrollpos.cury = 0;
        scrollpos.curytileoffs = 0;
    }
    if( scrollpos.curx*scrollpos.tilewidth+scrollpos.curxtileoffs+scrollvec.x >= 
            scrollpos.maxx*scrollpos.tilewidth+scrollpos.maxxtileoffs && 
            scrollvec.x > 0){
        scrollvec.t = 0;
        scrollvec.x = 0;
        scrollpos.curx = scrollpos.maxx;
        scrollpos.curxtileoffs = scrollpos.maxxtileoffs;
    }
    if( scrollpos.cury*scrollpos.tilewidth+scrollpos.curytileoffs+scrollvec.y >= 
	    scrollpos.maxy*scrollpos.tilewidth+scrollpos.maxytileoffs && 
	    scrollvec.y > 0){
        scrollvec.t = 0;
        scrollvec.y = 0;
        scrollpos.cury = scrollpos.maxy;
        scrollpos.curytileoffs = scrollpos.maxytileoffs;
    }
    
    if ((scrollvec.x == 0) && (scrollvec.y == 0)) {
        toscroll = false;
    	setValidScroll();
        return;
    }
    xtile = scrollpos.curxtileoffs+scrollvec.x;
    while( xtile < 0 ){
        scrollpos.curx--;
        xtile += scrollpos.tilewidth;
    }
    while( xtile >= scrollpos.tilewidth ){
        scrollpos.curx++;
        xtile -= scrollpos.tilewidth;
    }
    scrollpos.curxtileoffs = xtile;

    ytile = scrollpos.curytileoffs+scrollvec.y;
    while( ytile < 0 ){
        scrollpos.cury--;
        ytile += scrollpos.tilewidth;
    }
    while( ytile >= scrollpos.tilewidth ){
        scrollpos.cury++;
        ytile -= scrollpos.tilewidth;
    }
    scrollpos.curytileoffs = ytile;


    ++scrollvec.t;
    /* scrolling continues at current rate for scrolltime
     * passes then decays quickly */
    if (scrollvec.t >= scrolltime) {
        scrollvec.x /=2;
        scrollvec.y /=2;
    }
    setValidScroll();
}



/** sets the maximum value the scroll can take on.
 * @param the maximum x scroll.
 * @param the maximum y scroll.
 */

void CnCMap::setMaxScroll( Uint32 x, Uint32 y, Uint32 xtile, Uint32 ytile, Uint32 tilew )
{
    scrollpos.maxx = 0;
    scrollpos.maxy = 0;
    scrollpos.maxxtileoffs = 0;
    scrollpos.maxytileoffs = 0;
    if( xtile > 0 ){
        x++;
	xtile = tilew-xtile;
    }
    if( ytile > 0 ){
        y++;
	ytile = tilew-ytile;
    }

    if( width > x ){
        scrollpos.maxx = width - x;
        scrollpos.maxxtileoffs = xtile;
    }
    if( height > y ){
        scrollpos.maxy = height - y;
        scrollpos.maxytileoffs = ytile;
    }

    scrollpos.tilewidth = tilew;

    if( scrollpos.curx > scrollpos.maxx ){
        scrollpos.curx = scrollpos.maxx;
        scrollpos.curxtileoffs = scrollpos.maxxtileoffs;
    } else if( scrollpos.curx == scrollpos.maxx &&
            scrollpos.curxtileoffs > scrollpos.maxxtileoffs ){
        scrollpos.curxtileoffs = scrollpos.maxxtileoffs;
    }

    if( scrollpos.cury > scrollpos.maxy ){
        scrollpos.cury = scrollpos.maxy;
        scrollpos.curytileoffs = scrollpos.maxytileoffs;
    } else if( scrollpos.cury == scrollpos.maxy &&
            scrollpos.curytileoffs > scrollpos.maxytileoffs ){
        scrollpos.curytileoffs = scrollpos.maxytileoffs;
    }

    if( scrollpos.curxtileoffs > scrollpos.tilewidth ){
        scrollpos.curxtileoffs = scrollpos.tilewidth;
    }
    if( scrollpos.curytileoffs > scrollpos.tilewidth ){
        scrollpos.curytileoffs = scrollpos.tilewidth;
    }

    setValidScroll();
}

void CnCMap::setValidScroll(){
    valscroll = SCROLL_UP|SCROLL_DOWN|SCROLL_LEFT|SCROLL_RIGHT;
    if( scrollpos.curx == 0 && scrollpos.curxtileoffs == 0 ){
	valscroll &= ~SCROLL_LEFT;
    }
    if( scrollpos.cury == 0 && scrollpos.curytileoffs == 0 ){
	valscroll &= ~SCROLL_UP;
    }
    if( scrollpos.curx == scrollpos.maxx && 
	    scrollpos.curxtileoffs == scrollpos.maxxtileoffs ){
	valscroll &= ~SCROLL_RIGHT;
    }
    if( scrollpos.cury == scrollpos.maxy && 
	    scrollpos.curytileoffs == scrollpos.maxytileoffs ){
	valscroll &= ~SCROLL_DOWN;
    }

}

Uint16 CnCMap::getCost(Uint16 pos)
{
    return getCost(pos,NULL);
}

Uint16 CnCMap::getCost(Uint16 pos,Unit* excpUn)
{
    Uint16 cost;

    if( !ppool->getLPlayer()->getMapVis()[pos] && 
	(excpUn == NULL || excpUn->getDist(pos)>1 )){
	return 10;
    }

    switch(terraintypes[pos]) {
    case TTYPE_ROCK:
    case TTYPE_TREE:
    case TTYPE_WATER:
    case TTYPE_WATER_BLOCKED:
    case TTYPE_OTHER_NONPASS:
        cost = 0xffff;
        break;
    default:
        cost = 1 + uspool->getTileCost(pos,excpUn);
        break;
    }
    return cost;
}

SDL_Surface* CnCMap::getMiniMap()
{
    SDL_Surface *cminitile;
    if(minimap == NULL) {
        minimap = SDL_CreateRGBSurface (SDL_SWSURFACE, width*6,height*6, 16, 0xff, 0xff, 0xff, 0);
        SDL_Rect *pos = new SDL_Rect();
        ImageProc *ip = new ImageProc();
        pos->w = 6;
        pos->h = 6;
        pos->x = 0;
        pos->y = 0;
        SDL_Surface *maptileOne = getMapTile(0);
        minimap->format->Rmask = maptileOne->format->Rmask;
        minimap->format->Gmask = maptileOne->format->Gmask;
        minimap->format->Bmask = maptileOne->format->Bmask;
        minimap->format->Amask = maptileOne->format->Amask;
        if( maptileOne->format->palette != NULL ) {
            SDL_SetColors(minimap, maptileOne->format->palette->colors, 0, maptileOne->format->palette->ncolors);
        }
        int lineCounter = 0;
        for(Uint32 i = 0;  i < (Uint32) width*height; i++, pos->x += 6, lineCounter++) {
            if(lineCounter == width) {
                pos->y += 6;
                pos->x =0;
                lineCounter = 0;
            }
            cminitile = ip->minimapScale(getMapTile(i));
            SDL_BlitSurface(cminitile, NULL, minimap, pos);
            SDL_FreeSurface(cminitile);
        }
        delete pos;
        delete ip;
    }
    return minimap;

}

void CnCMap::storeLocation(Uint8 loc)
{
    if (loc >= NUMMARKS) {
        return;
    }
    scrollbookmarks[loc].x = scrollpos.curx;
    scrollbookmarks[loc].y = scrollpos.cury;
    scrollbookmarks[loc].xtile = scrollpos.curxtileoffs;
    scrollbookmarks[loc].ytile = scrollpos.curytileoffs;
}

void CnCMap::restoreLocation(Uint8 loc)
{
    if (loc >= NUMMARKS) {
        return;
    }
    scrollpos.curx = scrollbookmarks[loc].x;
    scrollpos.cury = scrollbookmarks[loc].y;
    scrollpos.curxtileoffs = scrollbookmarks[loc].xtile;
    scrollpos.curytileoffs = scrollbookmarks[loc].ytile;

    if( scrollpos.curxtileoffs >= scrollpos.tilewidth ){
	scrollpos.curxtileoffs = scrollpos.tilewidth-1;
    }
    if( scrollpos.curytileoffs >= scrollpos.tilewidth ){
	scrollpos.curytileoffs = scrollpos.tilewidth-1;
    }
    if( scrollpos.curx > scrollpos.maxx ){
	scrollpos.curx = scrollpos.maxx;
	scrollpos.curxtileoffs = scrollpos.maxxtileoffs;
    }else if( scrollpos.curx == scrollpos.maxx && 
	    scrollpos.curxtileoffs > scrollpos.maxxtileoffs ){
	scrollpos.curxtileoffs = scrollpos.maxxtileoffs;
    }
    if( scrollpos.cury > scrollpos.maxy ){
	scrollpos.cury = scrollpos.maxy;
	scrollpos.curytileoffs = scrollpos.maxytileoffs;
    }else if( scrollpos.cury == scrollpos.maxy && 
	    scrollpos.curytileoffs > scrollpos.maxytileoffs ){
	scrollpos.curytileoffs = scrollpos.maxytileoffs;
    }

    setValidScroll();
}

void CnCMap::printPlayersUnitsStructures()
{ // only for development purposes
    Uint16 i;
    std::vector<Structure*> structs;
    std::vector<Unit*> units;
    // for each player
    for (i = 0; i < ppool->getNumPlayers(); i++) {
        Player *p = ppool->getPlayer(i);
        std::cerr << "Player " << (int)i << " Name " << p->getName()
        << " Side " << p->getSide() << " UnitpalNum " << (int)p->getUnitpalNum()
        << endl << "  Power " << p->getPower() << " of PowerUsed " << p->getPowerUsed() << endl;
        // for all its units
        if (p->isDefeated()) {
            std::cerr << "Player has been defeated\n";
            continue;
        }
        units = p->getUnits();
        for (i=0; i < units.size(); ++i) {
            UnitType* utype = (UnitType *)units[i]->getType();
            std::cerr << "  Unit at Pos " << units[i]->getPos() << " Subpos "
            << units[i]->getSubpos() << " Health " << units[i]->getHealth()
            << " of MaxHealth " << (int)utype->getMaxHealth()
            << " Cost " << (int)utype->getCost() << endl
            << "    Unitnum " << units[i]->getUnitnum()
            << " isInfantry " << (int)utype->isInfantry() << endl
            << "    Speed " << (int)utype->getSpeed()
            << " Turnspeed " << (int)utype->getTurnspeed()
            << " Sight " << (int)utype->getSight()
            << " ROT " << (int)utype->getROT()
            << " MoveMod " << (int)utype->getMoveMod()
            << " TurnMod " << (int)utype->getTurnMod() << endl;
        }
        // for all its structures
        structs = p->getStructures();
        for (i = 0;i<structs.size();++i) {
            StructureType* stype = (StructureType *)structs[i]->getType();
            std::cerr << "  Structure at Pos " << structs[i]->getPos()
            << " Subpos " << structs[i]->getSubpos()
            << " Health " << structs[i]->getHealth()
            << " of MaxHealth " << (int)stype->getMaxHealth()
            << " Cost " << (int)stype->getCost() << endl
            << "    isWall " << (int)stype->isWall()
            << " Turnspeed " << (int)stype->getTurnspeed()
            << " Sight " << (int)stype->getSight() << endl;
        }
    }
}

Uint32 CnCMap::translateToPos(Uint16 x, Uint16 y){
    return y*width+x;
}

void CnCMap::translateFromPos(Uint32 pos, Uint16 *x, Uint16 *y){
    *y = pos/width;
    *x = pos-((*y)*width);
}
