// mode: -*- C++ -*-
/*****************************************************************************
 * map.h - header file for map class
 *
 ****************************************************************************/

#ifndef __MAP_H
#define __MAP_H


#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif


#include <vector>
#include <map>

class CnCMap;
class Unit;

#include "inifile.h"
#include "actioneventqueue.h"
#include "shpimage.h"
#include "unitandstructurepool.h"
#include "playerpool.h"
#include "timeline.h"
#include "shpimage.h"

#define HAS_OVERLAY 0x100
#define HAS_TERRAIN 0x200

/// number of map bookmarks
#define NUMMARKS 5

class LoadingScreen;
class NetConnection;

class TerrainEntry
{
public:
    Uint32 shpnum;
    Sint16 xoffset;
    Sint16 yoffset;
};

typedef struct
{
    Uint16 templateNum;
    Uint8 tileNum;
}
TileList;

typedef struct
{
    Sint8 x, y;
    Uint8 t;
}
ScrollVector;

typedef struct
{
    Uint16 x,y;
    Uint16 xtile,ytile;
}
ScrollBookmark;

typedef struct{
    Uint16 maxx, maxy;
    Uint16 curx, cury;
    Uint16 maxxtileoffs, maxytileoffs;
    Uint16 curxtileoffs, curytileoffs;
    Uint16 tilewidth;
}ScrollData;

class CnCMap
{
public:
    CnCMap(ActionEventQueue *evQueue, NetConnection *connection, LoadingScreen *lscreen);
    ~CnCMap();

    SDL_Surface *getMapTile( Uint32 pos )
    {
        return tileimages[tilematrix[pos]];
    }
    SDL_Surface *getShadowTile(Uint8 shadownum){
	if( shadownum >= numShadowImg ){
	    return NULL;
	}
	return shadowimages[shadownum];
    }
    Uint32 getTiberium(Uint32 pos)
    {
        return (overlaymatrix[pos] & 0xF);
    }
    Uint32 getSmudge(Uint32 pos)
    {
        return ((overlaymatrix[pos] & 0xF0) << 12);
    }
    Uint32 setSmudge(Uint32 pos, Uint8 value)
    {
        /// clear the existing smudge bits first
        overlaymatrix[pos] &= ~(0xF0);
        return (overlaymatrix[pos] |= (value<<4));
    }
    Uint32 setTiberium(Uint32 pos, Uint8 value)
    {
        /// clear the existing tiberium bits first
        overlaymatrix[pos] &= ~0xF;
        return (overlaymatrix[pos] |= value);
    }
    Uint32 getOverlay(Uint32 pos);
    Uint32 getTerrain(Uint32 pos, Sint16 &xoff, Sint16 &yoff);
    Uint8 getTerrainType(Uint32 pos){return terraintypes[pos];}

    std::vector<SHPImage *> *getImagePool();

    bool accScroll(Uint8 direction);
    bool absScroll(Sint16 dx, Sint16 dy);
    void doscroll();
    void setMaxScroll(Uint32 x, Uint32 y, Uint32 xtile, Uint32 ytile, Uint32 tilew);
    void setValidScroll();

    const char *getTheme()
    {
        return theme;
    }
    const char *getBrief()
    {
        return brief;
    }
    const char *getAction()
    {
        return action;
    }
    const char *getTheater()
    {
        return theater;
    }
    const char *getWinMov()
    {
        return winmov;
    }
    const char *getLoseMov()
    {
        return losemov;
    }
    UnitAndStructurePool *getUASP()
    {
        return uspool;
    }
    Uint32 getScrollPos()
    {
        return scrollpos.cury*width+scrollpos.curx;
    }
    Uint16 getXScroll()
    {
        return scrollpos.curx;
    }
    Uint16 getYScroll()
    {
        return scrollpos.cury;
    }
    Uint16 getXTileScroll(){
	return (Uint16)scrollpos.curxtileoffs;
    }
    Uint16 getYTileScroll(){
	return (Uint16)scrollpos.curytileoffs;
    }
    Uint16 getWidth()
    {
        return width;
    }
    Uint16 getX()
    {
        return x;
    }
    Uint16 getY()
    {
        return y;
    }
    Uint16 getHeight()
    {
        return height;
    }
    Uint8 getBuildlevel()
    {
        return buildlevel;
    }
    // const char *getTheater(){return theater;}

    Uint16 getCost(Uint16 pos);
    Uint16 getCost(Uint16 pos,Unit* excpUn);
    Uint8 getGameMode()
    {
        return gamemode;
    }
    PlayerPool *getPlayerPool()
    {
        return ppool;
    }
    NetConnection *getNetConnection()
    {
	return connection;
    }
    SDL_Surface* getMiniMap();
    bool toScroll()
    {
        return toscroll;
    }
    void storeLocation(Uint8 loc);
    void restoreLocation(Uint8 loc);
    void printPlayersUnitsStructures(); // only for development purposes
    TimeLine *getTimeLine()
    {
        return timeline;
    }
    std::vector<Uint16> getWaypoints()
    {
        return waypoints;
    }
    SHPImage* getPips()
    {
        return pips;
    }
    Uint32 getPipsNum()
    {
        return pipsnum;
    }
    SHPImage* getMoveFlash()
    {
        return moveflash;
    }
    Uint32 getMoveFlashNum()
    {
        return flashnum;
    }
    Uint32 translateToPos(Uint16 x, Uint16 y);
    void translateFromPos(Uint32 pos, Uint16 *x, Uint16 *y);
private:

    /* private functions used to load the map, these are in loadmap.cpp */
    /// load the ini part of the map
    void loadIni();
    /// the map section of the ini
    void simpleSections(INIFile *inifile);
    /// the advanced section of the ini
    void advancedSections(INIFile *inifile);
    /// load the bin part of the map (TD)
    void loadBin();
    /// extract RA map data
    void unMapPack( INIFile* inifile);
    /// load the palette
    void loadPal( SDL_Color *palette );
    /// read the bin part of the map (RA or TD)
    void parseBin(TileList *bindata);
    /// load a specified tile
    SDL_Surface *loadTile( INIFile *templini,
                           Uint16 templ, Uint8 tile, Uint32 &tiletype );

    /// width of map
    Uint16 width;
    /// height of map
    Uint16 height;
    /// current xpos
    //Uint16 xpos;
    /// current ypos
    //Uint16 ypos;
    /// x coordinate for the first tile
    Uint16 x;
    /// y coordinate for the first tile
    Uint16 y;

    ScrollData scrollpos;
    
    /// used to determine what units/structures can be built
    Uint8 buildlevel;

    /// the theater of the map (eg. winter)
    char *theater;
    /// the briefing movie
    char *brief;
    /// the actionmovie
    char *action;
    /// the players side (goodguy, badguy etc.)
    char *player;
    /// specific music to play for this mission
    char *theme;
    /// movie played after completed mission
    char *winmov;
    /// movie played after failed mission
    char *losemov;

    /// the name of the map
    char mapname[9];

    /* A array of tiles and a vector constining the images for the tiles */
    /// a matrix of the tiles
    Uint16 *tilematrix;
    std::vector<SDL_Surface *>tileimages;
    Uint16 numShadowImg;
    SDL_Surface **shadowimages;
    
    std::vector<Uint16> waypoints;

    Uint16 *overlaymatrix;
    Uint8 *terraintypes;

    std::vector<SHPImage *> imagepool;
    std::map<Uint32, TerrainEntry> terrains;
    std::map<Uint32, Uint16> overlays;

    Uint8 maptype; // 0: TD 1: RA
    Uint8 gamemode; // 0 - single player, 1 - skirmish, 2 - multiplayer

    SDL_Surface *minimap;

    ActionEventQueue *evQueue;

    ScrollVector scrollvec;
    bool toscroll;
    /// stores which directions can be scrolled
    Uint8 valscroll;
    /// stores certain map locations
    ScrollBookmark scrollbookmarks[NUMMARKS];
    Uint8 scrollstep, maxscroll, scrolltime;

    UnitAndStructurePool *uspool;
    PlayerPool* ppool;
    NetConnection *connection;
    TimeLine* timeline;

    Uint32 pipsnum;
    SHPImage* pips;
    Uint32 flashnum;
    SHPImage* moveflash;
};

class LoadMapError
    {}
;

#define SCROLL_NONE 0
#define SCROLL_UP 0x1
#define SCROLL_RIGHT 0x2
#define SCROLL_DOWN 0x4
#define SCROLL_LEFT 0x8
#define SCROLL_UPRIGHT (SCROLL_UP | SCROLL_RIGHT)
#define SCROLL_DOWNRIGHT (SCROLL_DOWN | SCROLL_RIGHT)
#define SCROLL_DOWNLEFT (SCROLL_DOWN | SCROLL_LEFT)
#define SCROLL_UPLEFT (SCROLL_UP | SCROLL_LEFT)

#define TTYPE_LAND 0
#define TTYPE_WATER 1
#define TTYPE_ROAD 2
#define TTYPE_ROCK 3
#define TTYPE_TREE 4
#define TTYPE_WATER_BLOCKED 5
#define TTYPE_OTHER_NONPASS 7


#endif

