// mode: -*- C++ -*-
#ifndef __SHPIMAGE_H
#define __SHPIMAGE_H

#include "SDL.h"
#include "imageproc.h"

class VFile;

typedef struct
{
    Uint16  NumImages;
    Uint16  Width;
    Uint16  Height;
    Uint32  *Offset;
    Uint8   *Format;
    Uint32  *RefOffs;
    Uint8   *RefFormat;
}
SHPHeader;

typedef struct
{
    Uint16 compression;
    Uint8 cy;
    Uint16 cx;
    Uint8 cy2;
    Uint16 size_in;
    Uint16 size_out;
}
Dune2Header;


class SHPBase
{
public:
    static void setPalette(SDL_Color *pal);
    static void calculatePalettes();
    static SDL_Color *getPalette(Uint8 palnum)
    {
        return palette[palnum];
    }
    SDL_Surface *scale(SDL_Surface *input, int quality)
    {
        return scaler.scale(input, quality);
    }
protected:
    static SDL_Color palette[32][256];
    int scaleq;
    ImageProc scaler;
};

class SHPImage : SHPBase
{
public:
    SHPImage(const char *fname, int scaleq );
    ~SHPImage();
    void getImage(Uint16 imgnum, SDL_Surface **img, SDL_Surface **shadow, Uint8 palnum);
    void getImageAsAlpha(Uint16 imgnum, SDL_Surface **img);
    Uint32 getWidth()
    {
        return header.Width;
    }
    Uint32 getHeight()
    {
        return header.Height;
    }
    Uint16 getNumImg()
    {
        return header.NumImages;
    }
private:
    void DecodeSprite(Uint8 *imgdst, Uint16 imgnum);
    static SDL_Color shadowpal[2];
    Uint8 *shpdata;
    SHPHeader header;
};

class Dune2Image : SHPBase
{
public:
    Dune2Image(const char *fname, int scaleq );
    ~Dune2Image();
    SDL_Surface *getImage(Uint16 imgnum);
private:
    Uint32 getD2Header(Uint16 imgnum);
    Uint8 *shpdata;
    Dune2Header header;
};

class TemplateImage : SHPBase
{
public:
    TemplateImage(const char *fname, int scaleq );
    TemplateImage(const char *fname, int scaleq, int ratemp );

    ~TemplateImage();
    Uint16 getNumTiles();
    SDL_Surface *getImage(Uint16 imgnum);
private:
    int ratemp;
    VFile *tmpfile;
};

class ImageNotFound
    {}
;

#endif
