// mode: -*- C++ -*-
#ifndef __STRUCTURE_H
#define __STRUCTURE_H

#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif

class Unit;
class UnitQueue;
class UnitOrStructure;
class Structure;
class StructureType;

#include <vector>
#include "SDL.h"
#include "shpimage.h"
#include "inifile.h"
#include "actioneventqueue.h"
#include "unitandstructurepool.h"
#include "unitorstructure.h"
#include "playerpool.h"
#include "ai.h"
#include "common.h"
#include "unitqueue.h"

class Weapon;
class WeaponsPool;

class StructureType : public UnitOrStructureType
{
public:
    StructureType(std::vector<SHPImage *>* imgPool, const char* type_name, INIFile *structini,
                  const char* thext,WeaponsPool* wpool);
    ~StructureType();
    Uint16 *getSHPNums()
    {
        return shpnums;
    }
    Uint16 *getSHPTNum()
    {
        return shptnum;
    }
    const char* getTName()
    {
        return tname;
    }
    const char* getName()
    {
        return name;
    }
    std::vector<char*> getPrereqs()
    {
        return prereqs;
    }
    std::vector<char*> getOwners()
    {
        return owners;
    }
    Uint8 getNumLayers()
    {
        return numshps;
    }

    Uint16 getMakeImg()
    {
        return makeimg;
    }
    static void setUASP(UnitAndStructurePool *usp)
    {
        uspool = usp;
    }
    bool isWall()
    {
        return is_wall;
    }

    Uint8 getXsize()
    {
        return xsize;
    }
    Uint8 getYsize()
    {
        return ysize;
    }

    Uint8 isBlocked(Uint16 tile)
    {
        return blocked[tile];
    }

    Sint8 getXoffset()
    {
        return xoffset;
    }
    Sint8 getYoffset()
    {
        return yoffset;
    }
    Uint8 getOffset()
    {
        return 0;
    }
    Uint16 getCost()
    {
        return cost;
    }
    Uint8 getTurnspeed()
    {
        return turnspeed;
    }
    Uint8 getSight()
    {
        return sight;
    }
    Uint8 getSpeed()
    {
        return 0;
    }
    armour_t getArmour()
    {
        return armour;
    }
    animinfo_t getAnimInfo()
    {
        return animinfo;
    }

    powerinfo_t getPowerInfo()
    {
        return powerinfo;
    }
    Uint16 getMaxHealth()
    {
        return maxhealth;
    }
    Weapon *getWeapon(bool primary = true)
    {
        return (primary?primary_weapon:secondary_weapon);
    }
    bool hasTurret()
    {
        return turret;
    }
    Uint16 getBlckOff()
    {
        return blckoff;
    }
    bool isInfantry()
    {
        return false;
    }
    Uint8 getNumWallLevels()
    {
        return numwalllevels;
    }
    Uint8 getDefaultFace()
    {
        return defaultface;
    }
    Uint8 getBuildlevel()
    {
        return buildlevel;
    }
    Uint8 getTechlevel()
    {
        return techlevel;
    }
    bool primarySettable()
    {
        return primarysettable;
    }
    Uint8 getMaxPassengers()
    {
        return maxpassengers;
    }
    std::vector<Uint8> getPassengerAllow()
    {
        return passenger_allow;
    }
    std::vector<UnitType*> getSpecificTypeAllow()
    {
        return specific_type_allow;
    }
private:
    Uint16 *shpnums, *shptnum;
    Uint16 cost,maxhealth,makeimg,blckoff;
    Sint8 xoffset,yoffset;
    armour_t armour;
    Uint8 turnspeed,sight,xsize,ysize,numshps,numwalllevels,defaultface;
    Uint8 techlevel,buildlevel,maxpassengers;
    Uint8 *blocked;
    char tname[8];
    char* name;
    std::vector<char*> owners;
    std::vector<char*> prereqs;

    Weapon* primary_weapon;
    Weapon* secondary_weapon;

    animinfo_t animinfo;
    powerinfo_t powerinfo;

    bool is_wall,turret, primarysettable;

    static UnitAndStructurePool* uspool;

    // matches the unit's type value specified in units.ini
    std::vector<Uint8> passenger_allow;
    // matches the unit's type name.
    std::vector<UnitType*> specific_type_allow;
};

class BAttackAnimEvent;

class Structure : public UnitOrStructure
{
public:
    friend class BuildingAnimEvent;
    friend class BAttackAnimEvent;

    Structure(StructureType *type, Uint16 cellpos, Uint8 owner);
    ~Structure();
    Uint8 getImageNums(Uint32 **inums, Sint8 **xoffsets, Sint8 **yoffsets, bool* showpips);
    Uint16* getImageNums()
    {
        return imagenumbers;
    }
    void changeImage(Uint8 layer, Sint16 imagechange)
    {
        imagenumbers[layer]+=imagechange;
    }
    Uint32 getImageNum(Uint8 layer)
    {
        return type->getSHPNums()[layer]+imagenumbers[layer];
    }
    void setImageNum(Uint32 num, Uint8 layer);
    UnitOrStructureType *getType()
    {
        return type;
    }
    void setStructnum(Uint16 stn)
    {
        structnum = stn;
    }
    Uint16 getStructnum()
    {
        return structnum;
    }
    Uint16 getPos()
    {
        return cellpos;
    }
    Uint16 getBPos(Uint16 curpos);
    void remove
        ();
    static void setUASP(UnitAndStructurePool *usp)
    {
        uspool = usp;
    }
    UnitAndStructurePool* getUSPool()
    {
        return uspool;
    }
    Uint16 getSubpos()
    {
        return 0;
    }
    void applyDamage(Sint16 amount, Weapon* weap, UnitOrStructure* attacker);
    void runAnim(Uint32 mode);
    void runSecAnim(Uint32 param);
    void stopAnim();
    void stop();
    Uint8 getOwner()
    {
        return owner;
    }
    void setOwner(Uint8 newowner)
    {
        owner = newowner;
    }
    bool canAttack()
    {
        return type->getWeapon()!=NULL;
    }
    void attack(UnitOrStructure* target);
    Uint16 getHealth()
    {
        return health;
    }
    Sint8 getXoffset()
    {
        return type->getXoffset();
    }
    Sint8 getYoffset()
    {
        return type->getYoffset();
    }
    bool isWall()
    {
        return type->isWall();
    }
    double getRatio()
    {
        return ratio;
    }
    bool isPrimary()
    {
        return primary;
    }
    void setPrimary(bool pri)
    {
        primary = pri;
    }
    std::vector<Unit*> getPassengers()
    {
        return passengers;
    }
    bool pushPassenger(Unit* un);
    Unit* removePassenger(Unit* un);
    bool canLoad(Unit* un);
    void unloadUnits();
    void loadUnits(vector<Unit*> units);
    void cancelLoad(Unit* un);
    bool isLoaded()
    {
        return false;
    }
    UnitQueue* getUnitQueue()
    {
        return unitqueue;
    }
    Uint32 getExitCell();
    LOADSTATE getLoadState()
    {
        return loadstate;
    }
    void resetLoadState(bool runsec, Uint32 param);
    Uint8 checkdamage();
    Uint16 getTargetCell();
private:
    StructureType *type;
    Uint16 *imagenumbers;
    Uint16 cellpos,bcellpos,health,structnum;
    Uint8 owner,references,damaged;
    bool animating,usemakeimgs,exploding,primary;
    double ratio; // health/maxhealth

    static UnitAndStructurePool* uspool;
    BuildingAnimEvent* buildAnim;
    BAttackAnimEvent* attackAnim;
    // used to store units that have been built and are about to leave
    // or units that have entered the building.
    // passengers is used for consistency with the Unit class.
    std::vector<Unit*> passengers;
    UnitQueue* unitqueue;
    LOADSTATE loadstate;
};

#include "structureanims.h"

#endif
