// mode: -*- C++ -*-
#ifndef __UNIT_H
#define __UNIT_H

#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif

#define MOVE_UP 0
#define MOVE_UPRIGHT 1
#define MOVE_RIGHT 2
#define MOVE_DOWNRIGHT 3
#define MOVE_DOWN 4
#define MOVE_DOWNLEFT 5
#define MOVE_LEFT 6
#define MOVE_UPLEFT 7


#include <vector>
#include <stack>
#include "SDL.h"
#include "shpimage.h"
#include "inifile.h"
#include "actioneventqueue.h"

class InfantryGroup;
class UnitType;
class Unit;
class UnitQueue;
class UnitOrStructure;
//class MoveAnimEvent;
#include "unitandstructurepool.h"
#include "unitanimations.h"
#include "path.h"
#include "unitorstructure.h"
#include "playerpool.h"
#include "ai.h"

#include "memtrack.h"
class Weapon;
class WeaponsPool;
class StructureType;
class L2Overlay;
class Talkback;

class UnitType : public UnitOrStructureType
{
public:
    UnitType(std::vector<SHPImage *> *imgPool, const char *type_name,
             INIFile *unitini, WeaponsPool *wpool);
    ~UnitType();
    Uint32 *getSHPNums()
    {
        return shpnums;
    }
    Uint8 getNumLayers()
    {
        return numlayers;
    }
    bool isInfantry()
    {
        return is_infantry;
    }
    Uint8 getType()
    {
        return unittype;
    }
    const char* getTName()
    {
        return tname;
    }
    const char* getName()
    {
        return name;
    }
    std::vector<char*> getPrereqs()
    {
        return prereqs;
    }
    std::vector<char*> getOwners()
    {
        return owners;
    }
    static void setUASP(UnitAndStructurePool *usp)
    {
        uspool = usp;
    }
    Uint8 getOffset()
    {
        return offset;
    }
    Uint8 getSpeed()
    {
        return speed;
    }
    Uint8 getROT()
    {
        return turnspeed;
    }
    Sint8 getMoveMod()
    {
        return movemod;
    }
    Uint8 getTurnMod()
    {
        return turnmod;
    }
    Uint16 getMaxHealth()
    {
        return maxhealth;
    }
    Uint16 getCost()
    {
        return cost;
    }
    Uint8 getTurnspeed()
    {
        return turnspeed;
    }
    Uint8 getSight()
    {
        return sight;
    }
    armour_t getArmour()
    {
        return armour;
    }
    const char* getRandTalk(TalkbackType type);
    Weapon *getWeapon(bool primary = true)
    {
        return (primary?primary_weapon:secondary_weapon);
    }
    bool isWall()
    {
        return false;
    }
    bool canDeploy()
    {
        return deployable;
    }
    const char* getDeployTarget()
    {
        return deploytarget;
    }
    StructureType* getDeployType()
    {
        return deploytype;
    }
    Uint8 getBuildlevel()
    {
        return buildlevel;
    }
    Uint8 getTechlevel()
    {
        return techlevel;
    }
    // what colour pip should be displayed for this unit when being carried
    Uint8 getPipColour()
    {
        return pipcolour;
    }
    Uint8 getMaxPassengers()
    {
        return maxpassengers;
    }
    vector<Uint8> getPassengerAllow()
    {
        return passenger_allow;
    }
    vector<UnitType*> getSpecificTypeAllow()
    {
        return specific_type_allow;
    }
private:
    Uint32 *shpnums;
    Uint16 cost,maxhealth;
    Uint8 numlayers,speed,turnspeed,turnmod,sight,offset,pipcolour;
    armour_t armour;
    Uint8 techlevel,buildlevel,unittype;
    Sint8 movemod;

    char tname[8];
    char* name;
    std::vector<char*> prereqs;
    std::vector<char*> owners;

    // Talkback related members
    Talkback* talkback;

    bool is_infantry, deployable;
    char* deploytarget;
    // this is used to check the unit can deploy
    StructureType* deploytype;
    Uint8 maxpassengers;
    // matches the unit's type value specified in units.ini
    vector<Uint8> passenger_allow;
    // matches the unit's type name.
    vector<UnitType*> specific_type_allow;

    Weapon* primary_weapon;
    Weapon* secondary_weapon;

    static UnitAndStructurePool* uspool;
};

class Unit : public UnitOrStructure
{
public:
    friend class UnitAnimEvent;
    friend class MoveAnimEvent;
    friend class UAttackAnimEvent;
    friend class TurnAnimEvent;
    Unit(UnitType *type, Uint16 cellpos, Uint8 subpos, InfantryGroup *group, Uint8 owner);
    ~Unit();
    Uint8 getImageNums(Uint32 **inums, Sint8 **xoffsets, Sint8 **yoffsets, bool* showpips);
    static void setUASP(UnitAndStructurePool *usp)
    {
        uspool = usp;
    }
    InfantryGroup *getInfantryGroup()
    {
        return infgrp;
    }
    void setInfantryGroup(InfantryGroup *ig)
    {
        infgrp = ig;
    }
    Uint32 getImageNum(Uint8 layer)
    {
        return type->getSHPNums()[layer]+imagenumbers[layer];
    }
    void setImageNum(Uint32 num, Uint8 layer);
    Sint8 getXoffset(); // return xoffset-type->getOffset();
    Sint8 getYoffset(); // return yoffset-type->getOffset();
    void setXoffset(Sint8 xo);
    void setYoffset(Sint8 yo);
    UnitOrStructureType* getType()
    {
        return type;
    }
    Uint16 getPos()
    {
        return cellpos;
    }
    Uint16 getBPos(Uint16 pos)
    {
        return cellpos;
    }
    Uint16 getSubpos()
    {
        return subpos;
    }
    Uint16 getUnitnum()
    {
        return unitnum;
    }
    void setUnitnum(Uint16 unum)
    {
        unitnum = unum;
    }
    Uint16 getHealth()
    {
        return health;
    }
    void move(Uint16 dest);
    void move(Uint16 dest, bool stop);
    void attack(UnitOrStructure* target);
    void attack(UnitOrStructure* target, bool stop);
    void turn(Uint8 facing, Uint8 layer);
    void stop();

    Uint8 getOwner()
    {
        return owner;
    }
    void setOwner(Uint8 newowner)
    {
        owner = newowner;
    }
    void remove
        ();
    void applyDamage(Sint16 amount, Weapon* weap, UnitOrStructure* attacker);
    bool canAttack()
    {
        return type->getWeapon()!=NULL;
    }
    void doRandTalk(TalkbackType ttype);
    UnitAndStructurePool* getUSPool()
    {
        return uspool;
    }
    void deploy(void);
    bool canDeploy();
    bool checkDeployTarget(Uint32 pos);
    Uint32 calcDeployPos();
    Uint32 getExitCell()
    {
        return calcDeployPos();
    }
    double getRatio()
    {
        return ratio;
    }
    bool pushPassenger(Unit* un);
    Unit* removePassenger(Unit* un);
    std::vector<Unit*> getPassengers()
    {
        return passengers;
    }
    bool canLoad(Unit* un);
    void unloadUnits();
    void loadUnits(std::vector<Unit*> units);
    UnitQueue* getUnitQueue()
    {
        return unitqueue;
    }
    bool isLoaded()
    {
        return (container != NULL);
    }
    void cancelLoad(Unit* un);
    void enterUnitOrStructure(UnitOrStructure* dest);
    void exitUnitOrStructure(UnitOrStructure* source);
    LOADSTATE getLoadState()
    {
        return loadstate;
    }
    void resetLoadState();
    Uint16 getDist(Uint16 pos);
    Uint16 getTargetCell();
private:
    UnitType *type;
    Uint16 *imagenumbers;
    Uint16 unitnum,cellpos,health,palettenum;
    Uint8 owner,references,subpos;
    Sint8 xoffset,yoffset;
    bool deployed;
    double ratio;

    L2Overlay* l2o;
    std::multimap<Uint16, L2Overlay*>::iterator l2entry;

    UnitOrStructure* container;
    std::vector<Unit*> passengers;
    UnitQueue* unitqueue;
    LOADSTATE loadstate;

    InfantryGroup *infgrp;

    MoveAnimEvent* moveanim;
    UAttackAnimEvent* attackanim;
    WalkAnimEvent* walkanim;
    TurnAnimEvent* turnanim1, *turnanim2;
    static UnitAndStructurePool *uspool;
};

/*
 * This should be a member of unit for infatry. When a infatry unit walks in
 * to a new cell a new group is created, if someone is already there that 
 * group is used. We need one more bit in the unit/structure matrix to tell if
 * a infantry is in that cell (faster). The groups should be destroyed 
 * (or reused by last unit if he travels to a empty cell) when the last
 * infantry leaves. */

class InfantryGroup
{
public:
    InfantryGroup()
    {
        int i;
        numinfantry = 0;
        for(i=0;i<5;i++)
            positions[i] = NULL;
    }
    void addInfantry(Unit *inf, Uint8 subpos)
    {
        positions[subpos] = inf;
        numinfantry++;
    }
    void removeInfantry(Uint8 subpos)
    {
        positions[subpos]=NULL;
        numinfantry--;
    }
    bool isClear(Uint8 subpos)
    {
        return (positions[subpos] == NULL);
    }
    Uint8 getNumInfantry()
    {
        return numinfantry;
    }
    Unit *unitAt(Uint8 subpos)
    {
        return positions[subpos];
    }
    Uint8 getImageNums(Uint32 **inums, Sint8 **xoffsets, Sint8 **yoffsets)
    {
        int i, j;

        (*inums) = new Uint32[numinfantry];
        (*xoffsets) = new Sint8[numinfantry];
        (*yoffsets) = new Sint8[numinfantry];
        j = 0;
        for( i = 0; i < 5; i++ ) {
            if( positions[i] != NULL ) {
                (*inums)[j]=positions[i]->getImageNum(0);
                (*xoffsets)[j]=positions[i]->getXoffset()+unitoffset[i];
                (*yoffsets)[j]=positions[i]->getYoffset()+unitoffset[i+5];
                j++;
            }
        }
        return numinfantry;
    }
    void getSubposOffsets(Uint8 oldsp, Uint8 newsp, Sint8* xoffs, Sint8* yoffs)
    {
        *xoffs = unitoffset[oldsp]-unitoffset[newsp];
        *yoffs = unitoffset[oldsp+5]-unitoffset[newsp+5];
    }
    static const Sint8* getunitoffsets()
    {
        return unitoffset;
    }
    Unit* getNearest(Uint8 subpos);
private:
    Unit* positions[5];
    Uint8 numinfantry;
    static const Sint8 unitoffset[];
    /* Theses values have been heavily tested, do NOT change them unless you're
     _really_ sure of what you are doing */
    /* X value */
    //      -13, -19, -7, -19, -7,
    /* Y value */
    //	-3, -7, -7, 1, 1
    //   };
};

class Talkback
{
public:
    Talkback(const char* talkback, INIFile* tbini);
    ~Talkback();
    const char* getRandTalk(TalkbackType type);
    static void setUASP(UnitAndStructurePool* usp) {
        uspool = usp;
    }
private:
    void merge(Talkback* mergee);
    std::vector<char*>* getTypeVector(TalkbackType type, bool ignoresize);
    TalkbackType getTypeNum(const char* name);
    std::vector<char*> tbReport;
    std::vector<char*> tbAck;
    std::vector<char*> tbDie;
    std::vector<char*> tbPostkill;
    std::vector<char*> tbAtkUn;
    std::vector<char*> tbAtkSt;
    static UnitAndStructurePool* uspool;
};

#endif
