// mode: -*- C++ -*-
#ifndef __UNITANIMATIONS_H
#define __UNITANIMATIONS_H


#include "SDL.h"
#include "path.h"

class Unit;
class UnitOrStructure;
class UnitAnimEvent;

class UnitAnimEvent : public ActionEvent
{
private:
    Unit* un;
    UnitAnimEvent* scheduled;
public:
    UnitAnimEvent(Uint32 p, Unit* un);
    virtual ~UnitAnimEvent();
    void setSchedule(UnitAnimEvent* e);
    void stopScheduled();
    virtual void stop() = 0;
    virtual void update() {}
    virtual void run() = 0;
};

class MoveAnimEvent : public UnitAnimEvent
{
private:
    bool stopping;
    void startMoveOne(bool wasblocked);
    void moveDone();
    Uint16 dest,newpos;
    bool blocked, moved_half, pathinvalid, waiting;
    Sint8 xmod, ymod;
    Unit* un;
    Path* path;
    Uint8 istep,dir;
    Uint32 range;
    MoveAnimEvent** storage;
public:
    MoveAnimEvent(Uint32 p, Unit* un, MoveAnimEvent** s);
    virtual ~MoveAnimEvent();
    virtual void stop();
    virtual void run();
    virtual void update();
    virtual void setRange(Uint32 nr)
    {
        range = nr;
    }
};

class WalkAnimEvent : public UnitAnimEvent
{
private:
    bool stopping;
    void calcbaseimage(void);
    Unit* un;
    Uint8 dir, istep, layer, baseimage;
    WalkAnimEvent** storage;
public:
    WalkAnimEvent(Uint32 p, Unit* un, Uint8 dir, Uint8 layer, WalkAnimEvent** s);
    virtual ~WalkAnimEvent();
    virtual void stop()
    {
        stopping = true;
    }
    virtual void run();
    virtual void changedir(Uint8 ndir)
    {
        stopping = false;
        dir = ndir;
        calcbaseimage();
    }
    void update() {}
};

class TurnAnimEvent : public UnitAnimEvent
{
private:
    bool stopping,runonce;
    Sint8 turnmod;
    Unit *un;
    Uint8 dir;
    Uint8 layer;
    TurnAnimEvent** storage;
public:
    TurnAnimEvent(Uint32 p, Unit *un, Uint8 dir, Uint8 layer, TurnAnimEvent** s);
    virtual ~TurnAnimEvent();
    virtual void run();
    virtual void stop()
    {
        stopping = true;
    }
    void update() {}
    virtual void changedir(Uint8 ndir)
    {
        stopping = false;
        dir = ndir;
    }
};

class UAttackAnimEvent : public UnitAnimEvent
{
private:
    Unit *un;
    bool stopping;
    Uint8 waiting;
    UnitOrStructure* target;
    UAttackAnimEvent** storage;
public:
    UAttackAnimEvent(Uint32 p, Unit *un, UAttackAnimEvent** s);
    virtual ~UAttackAnimEvent();
    void stop();
    virtual void update();
    virtual void run();
};

#endif /*__UNITANIMATIONS_H*/
