// mode: -*- C++ -*-
#ifndef __VFS_H
#define __VFS_H

#include "SDL.h"
#include "../vfs/archive.h"

/* Virtual file class.
 * All virtual files are opened for reading only since it's supposed to be 
 * used for graphics, sound and other data loading to games.
 */
class VFile
{
public:
    VFile(Uint32 filenum, Archive *arch)
    {
        this->filenum = filenum;
        this->archive = arch;
    }
    ~VFile()
    {
        archive->releaseFile(filenum);
    }
    Uint32 readByte(Uint8 *databuf, Uint32 numBytes)
    {
        return archive->readByte(filenum, databuf, numBytes);
    }
    Uint32 readWord(Uint16 *databuf, Uint32 numWords)
    {
        return archive->readWord(filenum, databuf, numWords);
    }
    Uint32 readThree(Uint32 *databuf, Uint32 numThrees)
    {
        return archive->readThree(filenum, databuf, numThrees);
    }
    Uint32 readDWord(Uint32 *databuf, Uint32 numDWords)
    {
        return archive->readDWord(filenum, databuf, numDWords);
    }
    char *getLine(char *string, Uint32 buflen)
    {
        return archive->readLine(filenum, string, buflen);
    }
    void seekSet(Uint32 pos)
    {
        archive->seekSet(filenum, pos);
    }
    void seekCur(Sint32 offset)
    {
        archive->seekCur(filenum, offset);
    }
    Uint32 fileSize()
    {
        return archive->getSize(filenum);
    }
    Uint32 tell()
    {
        return archive->getPos(filenum);
    }

    const char *getPath()
    {
        return archive->getPath(filenum);
    }
    //Archive *getArchive(){return archive;}
    //Uint32 getFileNum(){return filenum;}
private:
    Uint32 filenum;
    Archive *archive;
};

extern void VFS_Init(char *binpath);
extern void VFS_Destroy();
extern VFile *VFS_Open(const char *fname);
extern void VFS_Close(VFile *file);

#endif
