// mode: -*- C++ -*-
#ifndef __WEAPONSPOOL_H
#define __WEAPONSPOOL_H

#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif

#include <string>
#include <vector>
#include <map>

#include "SDL.h"
#include "common.h"

class INIFile;
class Sound;
class SHPImage;
class Unit;
class UnitOrStructure;
class Structure;
class UnitAndStructurePool;

class WeaponsPool;

class Warhead
{
public:
    Warhead(const char *whname, INIFile *weapini, WeaponsPool *wp);
    ~Warhead();
    //void getExplosion(Uint32 &image, Uint8 &steps){image = explosionimage; steps = explosionanimsteps;}
    Uint32 getEImage()
    {
        return explosionimage;
    }
    Uint8 getESteps()
    {
        return explosionanimsteps;
    }
    const char *getExplosionsound()
    {
        return explosionsound;
    }
    bool getWall()
    {
        return walls;
    }
    Uint8 getVersus(armour_t armour)
    {
        return versus[(Uint8)armour];
    }
private:
    //Uint8 explosiontype;
    Uint32 explosionimage;
    Uint8 explosionanimsteps;
    char *explosionsound;
    Uint8 infantrydeath;
    Uint8 blastradius;
    unsigned int versus[5];
    bool walls;
    bool trees;
    //Uint16 damage;
}
;

class Projectile
{
public:
    Projectile(const char *pname, INIFile *weapini, WeaponsPool *wp);
    ~Projectile();
    Uint32 getImageNum()
    {
        return imagenum;
    }
    //Uint8 getSpeed(){return speed;}
    bool doesRotate()
    {
        return rotates;
    }
private:
    Uint32 imagenum;
    Uint8 rotationimgs;
    bool AA;
    bool AG;
    //Uint8 speed;
    bool high, inacurate, rotates;
};

class Weapon
{
public:
    Weapon(const char *wname, WeaponsPool *wp);
    ~Weapon();
    Uint8 getReloadTime()
    {
        return reloadtime;
    }
    Uint8 getRange()
    {
        return range;
    }
    Uint8 getSpeed()
    {
        return speed;
    }
    Sint16 getDamage()
    {
        return damage;
    }
    bool getWall()
    {
        return whead->getWall();
    }
    // const char *getSound(){return firesound;}
    Projectile *getProjectile()
    {
        return projectile;
    }
    Warhead *getWarhead()
    {
        return whead;
    }
    void fire(UnitOrStructure* owner, Uint16 target, Uint8 subtarget, UnitAndStructurePool *uspool);
    //Uint32 tmppif;
    bool isHeatseek()
    {
        return heatseek;
    }
    bool isInaccurate()
    {
        return inaccurate;
    }
    WeaponsPool* getWeaponsPool()
    {
        return wp;
    }
    double getVersus(armour_t armour)
    {
        return (whead->getVersus(armour))/(double)100.0;
    }
private:
    Projectile *projectile;
    Warhead *whead;
    WeaponsPool* wp;
    Uint8 speed;
    Uint8 range;
    Uint8 reloadtime;
    Sint16 damage;
    Uint8 burst;
    bool heatseek,inaccurate;
    Uint32 fireimage;
    Uint32* fireimages;
    Uint8 numfireimages,numfiredirections;
    char *firesound;
};

class WeaponsPool
{
public:
    friend class Weapon;
    friend class Projectile;
    friend class Warhead;
    WeaponsPool(std::vector<SHPImage *> *imgpool);
    ~WeaponsPool();
    Weapon *getWeapon(const char *wname);
    INIFile* getWeaponsINI()
    {
        return weapini;
    }
private:
    std::vector<SHPImage *> *imgpool;
    std::map<std::string, Weapon*> weaponspool;
    std::map<std::string, Projectile*> projectilepool;
    std::map<std::string, Warhead*> warheadpool;
    INIFile* weapini;
};

#endif
