/*****************************************************************************
 * input.cpp - Class to handle all input from the user 
 *
 ****************************************************************************/

#include "input.h"
#include "unit.h"
#include "structure.h"
#include "soundengine.h"
#include "ai.h"
#include "projectileanim.h"
#include <string>
#include "logger.h"
#include "common.h"
#include "playerpool.h"
#include "config.h"
#include "memtrack.h"

bool Input::drawing;
SDL_Rect Input::markrect;
bool Input::minimapEnabled;


/** Constructor, sets up the input handeler.
 * @param the sidebar.
 * @param the map.
 * @param the width of the screen.
 * @param the height of the screen.
 */
Input::Input(Sidebar *sbar, CnCMap *map, Uint16 screenwidth, Uint16 screenheight,
             SDL_Rect *maparea, Cursor *cur, SoundEngine *sndengine)
{
    this->map = map;
    this->sidebar = sbar;
    width = screenwidth;
    height = screenheight;
    done = 0;
    donecount = 0;
    gamemode = map->getGameMode();
    finaldelay = getConfig().finaldelay;
    this->maparea = maparea;
    tabwidth = sbar->getTabLocation()->w;
    tilewidth = map->getMapTile(0)->w;
    this->sndengine = sndengine;
    drawing = false;
    selected = new Selection();
    kbdmod = 0;
    lplayer = map->getPlayerPool()->getLPlayer();
    lmousedown = MDOWN_NONE;
    rmousedown = MDOWN_NONE;

    cursor = cur;
    deploying_super = false;
    placing_struct = false;
    minimapEnabled = false;
    rcd_scrolling = false;
    sx = sy = 0;
    if (lplayer->isDefeated()) {
        logger->gameMsg("TEMPORARY FEATURE: Free MCV because of no initial");
        logger->gameMsg("units or structures.  Fixing involves adding triggers.");
        logger->gameMsg("Don't right click :-)");
        placing_struct = true;
        deploying_super = false;
        placeposvalid = true;
        temporary_place_unit = true;
        strcpy(placename,"MCV");
        map->getPlayerPool()->getLPlayer()->setVisBuild(SOB_BUILD,true);
        map->getPlayerPool()->getLPlayer()->setVisBuild(SOB_SIGHT, true);
        placetype = map->getUASP()->getStructureTypeByName("sbag");
    }
}

Input::~Input()
{
    delete selected;
}

/** Method to handle the input events. */
void Input::handle()
{
    SDL_Event event;
    int mx, my;
    Uint8 sdir, radarstat;
    Uint8* keystate;
    static ConfigType config = getConfig();

    while ( SDL_PollEvent(&event) ) {
        switch (event.type) {
        case SDL_MOUSEBUTTONDOWN:
            /** @bug These constants will change in SDL 1.3
             */
            // Mousewheel scroll up
            if (event.button.button == 4) {
                sidebar->scrollSidebar(true);
            // Mousewheel scroll down
            } else if (event.button.button == 5) {
                sidebar->scrollSidebar(false);
            } else {
                mx = event.button.x;
                my = event.button.y;
                if( mx >= maparea->x && mx < maparea->x + maparea->w &&
                        my >= maparea->y && my < maparea->y + maparea->h ) {
                    if( !rcd_scrolling && (event.button.button == SDL_BUTTON_LEFT )) { 
                        /* start drawing the selection square if in map */
                        lmousedown = MDOWN_MAP;
                        markrect.x = mx;
                        markrect.y = my;
                    } else if( event.button.button == SDL_BUTTON_RIGHT ) {
                        rmousedown = MDOWN_MAP;
                        sx = mx;
                        sy = my;
                        rcd_scrolling = true;
                    }
                }
                if (mx > width-tabwidth && sidebar->getVisible()) {
                    clickSidebar(mx,my);
                }
            }
            break;
        case SDL_MOUSEBUTTONUP:
            //	 SDL_GetMouseState(&mx, &my);
            mx = event.button.x;
            my = event.button.y;
            if (event.button.button == SDL_BUTTON_LEFT) {
                sidebar->resetButton();
                if( drawing ) {
                    selectRegion();
                } else if( my >= maparea->y ) {
                    if( my < maparea->y + maparea->h ) { /* map */
                        if( mx >= maparea->x && mx < maparea->x+maparea->w ) {
                            if( lmousedown == MDOWN_MAP) {
                                clickMap(mx, my);
                            }
                        }
                    }
                } else if( my < sidebar->getTabLocation()->h ){ /* clicked a tab */
                    if( mx < tabwidth ) {
                        //printf("Options menu\n");
                    } else if( mx > width-tabwidth ) {
                        sidebar->toggleVisible();
                    }
                }
                /* should also check if minimap is clicked, */

                /* We can't be drawing now */
                drawing = false;
                lmousedown = MDOWN_NONE;
            } else if (event.button.button == SDL_BUTTON_RIGHT) {
                // not moved mouse
                if ((mx < (sx+10)) && ((mx+10) > sx) && (my < (sy+10))  && ((my+10) > sy)) {
                    deploying_super = false;
                    placing_struct = false;
                    selected->clearSelection();
                } else {
                }
                rmousedown = MDOWN_NONE;
                rcd_scrolling = false;
            } 
            break;
            /* A key has been pressed or released */
        case SDL_KEYDOWN:
            /* If it wasn't a press, ignore this event */
            if( event.key.state != SDL_PRESSED )
                break;
            if (event.key.keysym.sym == config.bindablekeys[KEY_SIDEBAR]) {
                sidebar->toggleVisible();
            } else if (event.key.keysym.sym == config.bindablekeys[KEY_STOP]) {
                selected->stop();
            } else if (event.key.keysym.sym == config.bindablekeys[KEY_ALLY]) {
                if (gamemode != 0) {
                    Player* lplayer, *tplayer;
                    Unit* tmpunit = selected->getRandomUnit();
                    Structure* tmpstruct = selected->getRandomStructure();
                    lplayer = map->getPlayerPool()->getLPlayer();
                    if (tmpunit != NULL) {
                        tplayer = map->getPlayerPool()->getPlayer(tmpunit->getOwner());
                    } else if (tmpstruct != NULL) {
                        tplayer = map->getPlayerPool()->getPlayer(tmpstruct->getOwner());
                    } else {
                        tplayer = NULL;
                    }
                    if (tplayer != NULL) {
                        if (lplayer->allyWithPlayer(tplayer)) {
                            logger->gameMsg("%s allied with player %s",lplayer->getName(),tplayer->getName());
                        } else {
                            if (lplayer->unallyWithPlayer(tplayer)) {
                                logger->gameMsg("%s declared war on player %s",lplayer->getName(),tplayer->getName());
                            } else {
                                // tried to unally with self
                            }
                        }
                    }
                }
            } else {
                /* Otherwise check which key was pressed */
                switch( event.key.keysym.sym ) {
                    case SDLK_ESCAPE:
                        /* It was esc so quit */
                        done = 1;
                        break;
                    case SDLK_RSHIFT:
                    case SDLK_LSHIFT:
                        kbdmod |= KBDMOD_SHIFT;
                        break;
                    case SDLK_RCTRL:
                    case SDLK_LCTRL:
                        kbdmod |= KBDMOD_CTRL;
                        break;
                    case SDLK_RALT:
                    case SDLK_LALT:
                        kbdmod |= KBDMOD_ALT;
                        break;
                    case SDLK_1:
                    case SDLK_2:
                    case SDLK_3:
                    case SDLK_4:
                    case SDLK_5:
                    case SDLK_6:
                    case SDLK_7:
                    case SDLK_8:
                    case SDLK_9:
                    case SDLK_0:
                        if (kbdmod == KBDMOD_CTRL) {
                            selected->saveSelection(event.key.keysym.sym-48);
                        logger->gameMsg("Saved group %i",event.key.keysym.sym-48);
                    } else if (kbdmod == 0) {
                        Unit* tmpunit;
                        selected->loadSelection(event.key.keysym.sym-48);
                        tmpunit = selected->getRandomUnit();
                        if (tmpunit != NULL) {
                            logger->gameMsg("Group %i selected",event.key.keysym.sym-48);
                            sndengine->queueSound(((UnitType *)tmpunit->getType())
                                     ->getRandTalk(TB_report));
                        }
                    }
                    break;
                case SDLK_F1:
                case SDLK_F2:
                case SDLK_F3:
                case SDLK_F4:
                case SDLK_F5:
                    if (kbdmod == KBDMOD_CTRL) {
                        map->storeLocation(event.key.keysym.sym-282);
                    } else if (kbdmod == 0) {
                        map->restoreLocation(event.key.keysym.sym-282);
                    }
                    break;
                case SDLK_F7:
                    logger->gameMsg("M A R K @ %i",SDL_GetTicks());
                    logger->debug("Mark placed at %i\n",SDL_GetTicks());
                    break;
                case SDLK_F8:
                    map->getUASP()->showMoves();
                    break;
                case SDLK_F9: // pander for testing AI
                    if ((kbdmod & KBDMOD_SHIFT) != KBDMOD_SHIFT) {
                        logger->note(">>  AI activated for human player  <<\n");
                        if (lplayer->getAI() == NULL) {
                            lplayer->createAI(map);
                        }
                        lplayer->getAI()->attackOtherSide();
                    } else {
                        logger->note(">>  AI activated for all non-human players  <<\n");
                        for(Uint8 i = 0; i < map->getPlayerPool()->getNumPlayers(); i++) {
                            Player *pl = map->getPlayerPool()->getPlayer(i);
                            if (pl == NULL) {
                                continue;
                            }
                            if (pl != lplayer) {
                                if (!pl->getAI()) {
                                    pl->createAI(map);
                                }
                                pl->getAI()->attackOtherSide();
                            }
                        }
                    }
                    break;
                case SDLK_F10: // pander for testing AI
                    break;
                case SDLK_F11: // pander for testing AI
                    /*	        cerr << ">>  Places units for human and non-human players and optimises AI  <<" << endl;
                    for(Uint8 i = 0;
                    i < map->getPlayerPool()->getNumPlayers();
                    i++) {
                    Player *pl = map->getPlayerPool()->getPlayer(i);
                    pl->getAI()->optimise();
                    }*/
                    break;
                case SDLK_F12: // pander for testing AI
                    if ((kbdmod & KBDMOD_SHIFT)) {
                        map->getTimeLine()->update(map->getPlayerPool());
                        map->getTimeLine()->plot();
                    } else {
                        map->printPlayersUnitsStructures();
                    }
                    break;
                case SDLK_v:
                    if (!map->getPlayerPool()->getLPlayer()->canSeeAll()) {
                        map->getPlayerPool()->getLPlayer()->setVisBuild(SOB_SIGHT, true);
                        logger->gameMsg("Map revealed");
                    }
                    break;
                case SDLK_c:
                    if (!map->getPlayerPool()->getLPlayer()->canBuildAny()) {
                        map->getPlayerPool()->getLPlayer()->setVisBuild(SOB_BUILD, true);
                        logger->gameMsg("Build (nearly) anywhere");
                    }
                    break;
                case SDLK_b:
                    if (!map->getPlayerPool()->getLPlayer()->canBuildAll()) {
                        map->getUASP()->preloadUnitAndStructures(98);
                        map->getPlayerPool()->getLPlayer()->enableBuildAll();
                        sidebar->updateSidebar();
                        logger->gameMsg("Prerequisites disabled");
                    }
                    break;
                default:
                    break;
                }
                break;
            }
        case SDL_KEYUP:
            if( event.key.state != SDL_RELEASED )
                break;

            switch( event.key.keysym.sym ) {
            case SDLK_RSHIFT:
            case SDLK_LSHIFT:
                kbdmod &= ~KBDMOD_SHIFT;
                break;
            case SDLK_RCTRL:
            case SDLK_LCTRL:
                kbdmod &= ~KBDMOD_CTRL;
                break;
            case SDLK_RALT:
            case SDLK_LALT:
                kbdmod &= ~KBDMOD_ALT;
                break;
            default:
                break;
            }
            break;

            /* We got the quit event so quit */
        case SDL_QUIT:
            done = 1;
            break;
            /* By default we just ignore the event */
        default:
            break;
        }
    }
    sdir = 0;
    keystate = SDL_GetKeyState(NULL);
    if (keystate[SDLK_LEFT])
        sdir |= SCROLL_LEFT;
    else if (keystate[SDLK_RIGHT])
        sdir |= SCROLL_RIGHT;
    if (keystate[SDLK_UP])
        sdir |= SCROLL_UP;
    else if (keystate[SDLK_DOWN])
        sdir |= SCROLL_DOWN;
    if (sdir != SCROLL_NONE)
        map->accScroll(sdir);

    if (keystate[SDLK_PAGEDOWN])
        sidebar->scrollSidebar(false);
    else if (keystate[SDLK_PAGEUP])
        sidebar->scrollSidebar(true);

    if( map->getUASP()->hasDeleted() ) {
        selected->checkSelection();
    }

    if (map->getPlayerPool()->pollSidebar()) {
        sidebar->updateSidebar();
    }
    radarstat = map->getPlayerPool()->statRadar();
    switch (radarstat) {
    case 0: // do nothing
        break;
    case 1: // got radar
        sidebar->startRadarAnim(0,&minimapEnabled);
        break;
    case 2: // lost radar
    case 3: // radar powered down
        minimapEnabled = false;
        sidebar->startRadarAnim(1,&minimapEnabled);
        break;
    default:
        logger->error("BUG: unexpected value returned from PlayerPool::statRadar: %i\n",
                      radarstat);
        break;
    }

    updateMousePos();

    if( map->getPlayerPool()->hasWon() || map->getPlayerPool()->hasLost() ) {
        ++donecount;
    }
    if (donecount == 1) {
        if (map->getPlayerPool()->hasWon()) {
            sndengine->queueSound("accom1.aud");
            logger->gameMsg("MISSION ACCOMPLISHED");
        } else {
            sndengine->queueSound("fail1.aud");
            logger->gameMsg("MISSION FAILED");
        }
    }
    if (donecount > finaldelay) {
        done = 1;
    }
}

void Input::updateMousePos()
{
    Uint8 scroll;
    Uint16 cursornum;
    /* Check the position of the mousepointer and scroll if we are
     * less than 10 pixels from an edge. */
    int mx, my;
    SDL_GetMouseState(&mx, &my);


    if( drawing ) { /* set cursor to the default one when drawing */
        if( mx < maparea->x )
            markrect.w = maparea->x;
        else if( mx >= maparea->x+maparea->w )
            markrect.w = maparea->x+maparea->w-1;
        else
            markrect.w = mx;
        if( my < maparea->y )
            markrect.h = my = maparea->y;
        else if( my >= maparea->y+maparea->h )
            markrect.h = maparea->y+maparea->h-1;
        else
            markrect.h = my;

        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
    }
    /* Check if we should start drawing */
    else if( lmousedown == MDOWN_MAP && !placing_struct && !deploying_super) {
        if( my >= maparea->y && my < maparea->y+maparea->h &&
                mx >= maparea->x && mx < maparea->x+maparea->w )
            if(abs(markrect.x - mx) > 2 || abs(markrect.y - my) > 2) {
                drawing = true;
                markrect.w = mx;
                markrect.h = my;
            }
    } else { /* set the correct cursor at the end of this else */
        scroll = 0;
        if (( mx > width - 10 ) || (rcd_scrolling && (mx > (sx+10))) )
            scroll |= SCROLL_RIGHT;
        else if (( mx < 10 ) || (rcd_scrolling && ((mx+10) < sx)) )
            scroll |= SCROLL_LEFT;
        if (( my > height - 10 ) || (rcd_scrolling && (my > (sy+10))) )
            scroll |= SCROLL_DOWN;
        else if (( my < 10 ) || (rcd_scrolling && ((my+10) < sy)) )
            scroll |= SCROLL_UP;

        if( scroll != 0 ) {
            switch(scroll) {
            case SCROLL_UP:
                cursornum = CUR_SCROLLUP;
                break;
            case SCROLL_UPLEFT:
                cursornum = CUR_SCROLLUL;
                break;
            case SCROLL_LEFT:
                cursornum = CUR_SCROLLLEFT;
                break;
            case SCROLL_DOWNLEFT:
                cursornum = CUR_SCROLLDL;
                break;
            case SCROLL_DOWN:
                cursornum = CUR_SCROLLDOWN;
                break;
            case SCROLL_DOWNRIGHT:
                cursornum = CUR_SCROLLDR;
                break;
            case SCROLL_RIGHT:
                cursornum = CUR_SCROLLRIGHT;
                break;
            case SCROLL_UPRIGHT:
                cursornum = CUR_SCROLLUR;
                break;
            default:
                cursornum = CUR_STANDARD;
            }

            if (rcd_scrolling) {
                if (!map->absScroll((mx-sx),(my-sy)))
                    cursornum += CUR_NOSCROLL_OFFSET;
            } else {
                if (!map->accScroll(scroll))
                    cursornum += CUR_NOSCROLL_OFFSET;
            }
            if (!deploying_super) {
                cursor->setCursor(cursornum, CUR_NOANIM);
            }
        } else if( placing_struct ) {
            checkPlace(mx, my);
            return;
        } else if (deploying_super) {
            cursor->setXY(mx, my);
            return;
        } else {
            setCursorNum(mx, my);
        }
    }
    cursor->setXY(mx, my);
}

void Input::clickMap(int mx, int my)
{
    Unit *curunit;
    Unit* tmpunit = NULL;
    InfantryGroup* ig;
    Structure *curstructure;
    Uint16 pos;
    Uint8 subpos;
    bool sndplayed = false;
    bool enemy;
    Player* lplayer = map->getPlayerPool()->getLPlayer();

    clickedTile(mx, my, &pos, &subpos);
    if( pos == 0xffff )
        return;

    if( placing_struct ) {
        pos = checkPlace(mx, my);
        if( pos != 0xffff ) {
            placing_struct = false;
            if( temporary_place_unit ) {
                if (lplayer->isDefeated()) {
                    lplayer->setVisBuild(SOB_BUILD, false);
                }
                map->getUASP()->createUnit(placename, pos, 0, lplayer->getPlayerNum());
            } else {
                map->getUASP()->createStructure(placename, pos, lplayer->getPlayerNum(),true);
                sndengine->queueSound("hvydoor1.aud");
                if(!( (strcasecmp("sbag",placename) == 0) ||
                        (strcasecmp("brik",placename) == 0) ||
                        (strcasecmp("cycl",placename) == 0) ||
                        (strcasecmp("wood",placename) == 0) ||
                        (strcasecmp("barb",placename) == 0))) {
                    sndengine->queueSound("constru2.aud");
                    sidebar->updateSidebar();
                }
            }
        }
        return;
    } else if (deploying_super) {
        deploying_super = false;
        if (strcmp("ion",placename) == 0) {
            sndengine->queueSound("ion1.aud");
        }
        return;
    }

    curunit = map->getUASP()->getUnitAt(pos, subpos);
    curstructure = map->getUASP()->getStructureAt(pos,selected->getWall());
    ig = map->getUASP()->getInfantryGroupAt(pos);
    if (ig != NULL && curunit == NULL) {
        curunit = ig->getNearest(subpos);
        if (curunit->getOwner() == map->getPlayerPool()->getLPlayerNum()) {
            curunit = NULL;
        }
    }
    if( curunit != NULL ) {
        enemy = curunit->getOwner() != map->getPlayerPool()->getLPlayerNum();
        if (enemy) {
            if (selected->canAttack() && (
                        !(lplayer->isAllied(
                              map->getPlayerPool()->getPlayer(curunit->getOwner()))
                         ) ||
                        kbdmod == KBDMOD_CTRL) ) {
                selected->attackUnit(curunit);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(TB_atkun));
                }
            } else {
                selected->clearSelection();
                selected->addUnit(curunit, enemy);
            }
            return;
        } else if( kbdmod == KBDMOD_CTRL ) {
            if( selected->canAttack() ) {
                selected->attackUnit(curunit);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(TB_atkun));
                }
            }
        } else if( kbdmod == KBDMOD_SHIFT ) {
            if( selected->isEnemy() ) {
                selected->clearSelection();
            }
            if( curunit->isSelected() )
                selected->removeUnit(curunit);
            else {
                selected->addUnit(curunit, false);
                if (!sndplayed) {
                    sndengine->queueSound(((UnitType *)curunit->getType())->getRandTalk(TB_report));
                    sndplayed = true;
                }
            }
        } else if( !curunit->isSelected() ) { 
            /*if (!enemy && !selected->empty() && !selected->isEnemy() && selected->canLoad(curunit)) {
              selected->loadUnits(curunit);
              } else { */
            selected->clearSelection();
            selected->addUnit(curunit, false);
            //}
            if (!sndplayed) {
                sndengine->queueSound(((UnitType *)curunit->getType())->getRandTalk(TB_report));
                sndplayed = true;
            }
        } else if ((!selected->isEnemy())&&(curunit->canDeploy())) {
            selected->removeUnit(curunit);
            selected->purgeUnit(curunit);
            curunit->deploy();
            sidebar->updateSidebar();
        }
    } else if( curstructure != NULL ) {
        enemy = curstructure->getOwner() != map->getPlayerPool()->getLPlayerNum();

        if( enemy ) {
            if (selected->canAttack() && (
                        !(lplayer->isAllied(
                              map->getPlayerPool()->getPlayer(curstructure->getOwner()))
                         ) ||
                        kbdmod == KBDMOD_CTRL) ) {
                selected->attackStructure(curstructure);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(TB_atkst));
                }
            } else if (!curstructure->isWall()) {
                selected->clearSelection();
                selected->addStructure(curstructure, enemy);
            }
            return;
        }
        if( kbdmod == KBDMOD_CTRL ) {
            if( selected->canAttack() ) {
                selected->attackStructure(curstructure);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(TB_atkst));
                }
                return;
            }
        } else if( kbdmod == KBDMOD_SHIFT ) {
            if( selected->isEnemy() ) {
                selected->clearSelection();
            }
            if( curstructure->isSelected() )
                selected->removeStructure(curstructure);
            else {
                if (!curstructure->isWall())
                    selected->addStructure(curstructure, false);
            }
            return;
        } else if( kbdmod == KBDMOD_ALT ) {
            // hack
            curstructure->runSecAnim(5);
            return;
        } else if( !curstructure->isSelected() ) {
            /*if (!enemy && !selected->empty() && !selected->isEnemy() && selected->canLoad(curstructure)) {
              selected->loadUnits(curstructure);
              } else {*/
            selected->clearSelection();
            if (!curstructure->isWall()) {
                selected->addStructure(curstructure, false);
            }
            //}
            return;
        } else {
            if ( (curstructure != lplayer->getPrimary((StructureType*)curstructure->getType())) &&
                    (((StructureType*)curstructure->getType())->primarySettable()) ) {
                lplayer->setPrimary(curstructure);
                sndengine->queueSound("pribldg1.aud");
            }
            return;
        }
    } else {
        if( selected->canMove() && map->getCost(pos) < 0xfff0) {
            if (!sndplayed) {
                sndengine->queueSound(((UnitType *)selected->getRandomUnit()->getType())->getRandTalk(TB_ack));
                sndplayed = true;
            }
            selected->moveUnits(pos);
            new ExplosionAnim(1, map->getUASP(), pos, map->getMoveFlashNum(),
                              map->getMoveFlash()->getNumImg(), 0, 0);
        }
    }
}

void Input::clickedTile(int mx, int my, Uint16* pos, Uint8* subpos)
{
    Uint16 xrest, yrest, tx, ty;
    mx -= maparea->x-map->getXTileScroll();
    my -= maparea->y-map->getYTileScroll();
    tx = mx/tilewidth;
    ty = my/tilewidth;
    if( tx >= map->getWidth() || ty >= map->getHeight() )
        *pos = 0xffff;
    else {
        *pos = (my/tilewidth)*map->getWidth()+mx/tilewidth;
        *pos += map->getScrollPos();
    }

    xrest = mx%tilewidth;
    yrest = my%tilewidth;
    *subpos = 0;
    if( xrest >= 8 && xrest < 16 && yrest >= 8 && yrest < 16 )
        return;

    *subpos = 1;
    if( yrest >= 12 )
        (*subpos)+=2;
    if( xrest >= 12 )
        (*subpos)++;
    /* assumes the subpositions are:
     * 1 2
     *  0
     * 3 4
     */
}

void Input::setCursorNum(int mx, int my)
{
    Uint16 pos;
    Uint8 subpos;
    Unit *curunit;
    Structure *curstruct;
    bool enemy;

    /* clicked the map */
    if( my >= maparea->y && my < maparea->y+maparea->h &&
            mx >= maparea->x && mx < maparea->x+maparea->w ) {
        clickedTile(mx, my, &pos, &subpos);
        if( pos != 0xffff ) {
            if( map->getPlayerPool()->getLPlayer()->getMapVis()[pos] ) {
                InfantryGroup* ig;
                curunit = map->getUASP()->getUnitAt(pos, subpos);
                curstruct = map->getUASP()->getStructureAt(pos,selected->getWall());
                ig = map->getUASP()->getInfantryGroupAt(pos);
                if (ig != NULL && curunit == NULL) {
                    curunit = ig->getNearest(subpos);
                    if (curunit->getOwner() == map->getPlayerPool()->getLPlayerNum()) {
                        curunit = NULL;
                    }
                }
            } else {
                curunit = NULL;
                curstruct = NULL;
            }
            if( curunit != NULL ) {
                enemy = !(lplayer->isAllied(map->getPlayerPool()->getPlayer(curunit->getOwner())));
                if( !curunit->isSelected() ) {
                    if( selected->canAttack() && (enemy || (kbdmod == KBDMOD_CTRL))) {
                        cursor->setCursor("attack");
                        return;
                    }
                    /*
                    if (!enemy && !selected->empty() && !selected->isEnemy() && selected->canLoad(curunit)) {
                    cursor->setCursor("enter");
                    return;
                    }*/
                    if( selected->empty() ||
                        (!selected->empty() && !selected->isEnemy() && kbdmod == KBDMOD_SHIFT) ||
                        (!selected->canAttack() && kbdmod != KBDMOD_CTRL) ) {
                        cursor->setCursor("select");
                        return;
                    } else {
                        cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                        return;
                    }
                } else {
                    if (curunit->getOwner() == map->getPlayerPool()->getLPlayerNum()) {
                        if (((UnitType*)curunit->getType())->canDeploy()) {
                            if (curunit->canDeploy())
                                cursor->setCursor("deploy");
                            else
                                cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                            return;
                        } else {
                            cursor->setCursor("select");
                            return;
                        }
                    }
                }
            } else if( curstruct != NULL ) {
                enemy = !(lplayer->isAllied(map->getPlayerPool()->getPlayer(curstruct->getOwner())));
                if( !curstruct->isSelected() ) {
                    if( selected->canAttack() && (enemy || (kbdmod == KBDMOD_CTRL))) {
                        cursor->setCursor("attack");
                        return;
                    } else if( !enemy || selected->empty() || selected->isEnemy()) {
                        if (curstruct->isWall()) {
                            cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                            return;
                        } else {
                            /*if (!enemy && !selected->empty() && !selected->isEnemy() && selected->canLoad(curstruct)) {
                              cursor->setCursor("enter");
                              return;
                              }*/
                            if( selected->empty() ||
                                (!selected->empty() && !selected->isEnemy()) ||
                                !selected->canAttack() ) {
                                cursor->setCursor("select");
                                return;
                            }
                        }
                    } else if (enemy) {
                        // since we have already considerred the case
                        // when we can attack, we can't attack here.
                        if (kbdmod == KBDMOD_CTRL) {
                            cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                        } else {
                            cursor->setCursor("select");
                        }
                        return;
                    }
                    return;
                } else {
                    if (!enemy) {
                        if ((curstruct != lplayer->getPrimary((StructureType*)curstruct->getType()))&&
                                (((StructureType*)curstruct->getType())->primarySettable()) ) {
                            cursor->setCursor("deploy");
                            return;
                        } else {
                            cursor->setCursor("select");
                            return;
                        }
                    } else {
                        cursor->setCursor("select");
                        return;
                    }
                }
            } else if( selected->canMove() ) {
                if( map->getCost(pos) < 0xfff0 || !map->getPlayerPool()->getLPlayer()->getMapVis()[pos]) {
                    cursor->setCursor(CUR_MOVE, CUR_NOANIM);
                } else {
                    cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                }
                return;
            }
        }
    }

    cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
}

void Input::selectRegion()
{
    Uint16 startx, starty, stopx, stopy;
    Uint16 scannerx, scannery, curpos, i;
    Unit *un;
    bool playedsnd = false;

    if (kbdmod != KBDMOD_SHIFT)
        selected->clearSelection();

    startx = (MIN(markrect.x, markrect.w)-maparea->x+map->getXTileScroll())/tilewidth+map->getXScroll();
    starty = (MIN(markrect.y, markrect.h)-maparea->y+map->getYTileScroll())/tilewidth+map->getYScroll();
    stopx = (abs(markrect.x - markrect.w)+1)/tilewidth+startx;
    stopy = (abs(markrect.y - markrect.h)+1)/tilewidth+starty;
    if( stopx >= map->getWidth() )
        stopx = map->getWidth() - 1;
    if( stopy >= map->getHeight() )
        stopy = map->getHeight() - 1;

    for( scannery = starty; scannery <= stopy; scannery++ ) {
        for( scannerx = startx; scannerx <= stopx; scannerx++ ) {
            curpos = scannery*map->getWidth()+scannerx;
            for( i = 0; i < 5; i++ ) {
                un = map->getUASP()->getUnitAt(curpos, i);
                if( un != NULL ) {
                    if( un->getOwner() != map->getPlayerPool()->getLPlayerNum() ) {
                        continue;
                    }
                    selected->addUnit(un, false);
                    if (!playedsnd) {
                        sndengine->queueSound(((UnitType *)un->getType())->getRandTalk(TB_report));
                        playedsnd = true;
                    }
                    if( !((UnitType *)un->getType())->isInfantry() )
                        break;
                }
            }
        }
    }

}

void Input::clickSidebar(int mx, int my)
{
    Uint8 createmode, butclick;

    mx -= (width-tabwidth);
    my -= sidebar->getTabLocation()->h;
    butclick = sidebar->getButton(mx,my);
    if (butclick != 255) {
        /** @todo find a more elegant way to do this, as scrolling will blank
         *  current place.
         */
        strncpy(placename,"xxxx",4);
        sidebar->clickButton(butclick,placename,&createmode);

        if( (strncasecmp("xxxx",placename,4) == 0) ) {
            placing_struct = false;
            deploying_super = false;
            return;
        }
        if( (strncasecmp("atom",placename,4) == 0)
                || (strncasecmp("ion",placename,3) == 0) ||(strncasecmp("bomb",placename,4) == 0)) {
            placing_struct = false;
            deploying_super = true;
            if ( (strncasecmp("atom",placename,4) == 0) || (strncasecmp("ion",placename,3) == 0)) {
                cursor->setCursor(placename);
            } else {
                cursor->setCursor("superdeploy");
            }
            sndengine->queueSound("select1.aud");
            return;// was scrolling or clicked on super weap
        }
        /* ugly++, place units as structures */
        if( createmode ) {
            placing_struct = true;
            deploying_super = false;
            placeposvalid = true;
            temporary_place_unit = true;
            placetype = map->getUASP()->getStructureTypeByName("sbag");
            return;
        }
        temporary_place_unit = false;
        placing_struct = true;
        deploying_super = false;
        placeposvalid = true;
        placetype = map->getUASP()->getStructureTypeByName(placename);
    }
}

Uint16 Input::checkPlace(int mx, int my)
{
    Uint8 placexpos, placeypos;
    Uint8 *placemat;
    bool* buildable;
    Uint16 pos, curpos, placeoff;
    Uint8 subpos;

    if( my < maparea->y || my >= maparea->y+maparea->h ||
            mx < maparea->x || mx >= maparea->x+maparea->w ) {
        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
        cursor->setXY(mx, my);
        return 0xffff;
    }

    clickedTile(mx, my, &pos, &subpos);
    if( pos == 0xffff ) {
        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
        cursor->setXY(mx, my);
        placeposvalid = false;
        return pos;
    }

    if( pos%map->getWidth() + placetype->getXsize() - map->getXScroll() > MIN(maparea->w/tilewidth, map->getWidth())) {
        pos -= (pos%map->getWidth() + placetype->getXsize() - map->getXScroll() - MIN(maparea->w/tilewidth, map->getWidth()));
    }
    if( pos/map->getWidth() + placetype->getYsize() - map->getYScroll() > MIN(maparea->h/tilewidth, map->getHeight())) {
        pos -= ((pos/map->getWidth() + placetype->getYsize() - map->getYScroll() - MIN(maparea->h/tilewidth, map->getHeight())))*map->getWidth();
    }

    buildable = lplayer->getMapBuildable();

    /* check if pos is valid and set cursor */
    placeposvalid = true;
    // note: assumes land based buildings for now
    map->getUASP()->setCostCalcOwnerAndType(lplayer->getPlayerNum(),0);
    placemat = new Uint8[placetype->getXsize()*placetype->getYsize()];
    for( placeypos = 0; placeypos < placetype->getYsize(); placeypos++ ) {
        for( placexpos = 0; placexpos < placetype->getXsize(); placexpos++ ) {
            curpos = pos+placeypos*map->getWidth()+placexpos;
            placeoff = placeypos*placetype->getXsize()+placexpos;
            if( placetype->isBlocked(placeoff) ) {
                placemat[placeoff] = 1;
                if( map->getCost(curpos)>1 || (!buildable[curpos]) ) {
                    placeposvalid = false;
                    placemat[placeoff] = 2;
                }
            } else {
                placemat[placeoff] = 0;
            }
        }
    }
    cursor->setPlaceCursor(placetype->getXsize(), placetype->getYsize(), placemat);

    delete[] placemat;
    drawing = false;

    mx = (pos%map->getWidth() - map->getXScroll())*tilewidth + maparea->x-map->getXTileScroll();
    my = (pos/map->getWidth() - map->getYScroll())*tilewidth + maparea->y-map->getYTileScroll();
    cursor->setXY(mx, my);
    if( !placeposvalid ) {
        pos = 0xffff;
    }
    return pos;
}
