#include <string.h>
#include "logger.h"
#include "message.h"
#include "memtrack.h"

Logger::Logger(const char *logname, int threshold)
{
    logfile = fopen(logname, "w");
    if( logfile == NULL ) {
        printf("Unable to open logfile \"%s\" for writing, using stdout instead\n", logname);
        logfile = stdout;
    }
    indentsteps = 0;
    indentString[0] = '\0';
    msgpool = NULL;
    rendergamemsg = false;
    this->threshold = threshold;
}

Logger::~Logger()
{
    fclose(logfile);
}

void Logger::indent()
{
    if( indentsteps > 60 ) {
        indentsteps = 62;
    } else {
        strcat(indentString, "  ");
        indentsteps+=2;
    }
    indentString[indentsteps] = '\0';
}

void Logger::unindent()
{
    if( indentsteps < 3 ) {
        indentsteps = 0;
    } else {
        indentsteps -= 2;
    }
    indentString[indentsteps] = '\0';
}

void Logger::error(const char *txt, ...)
{
    va_list ap;
    fprintf(logfile, "%sERROR: ", indentString);
    va_start(ap, txt);
    vfprintf(logfile, txt, ap);
#if !defined _WIN32
    printf("%sERROR: ", indentString);
    vprintf(txt, ap);
    fflush(stdout);
#endif
    va_end(ap);
    fflush(logfile);
}

void Logger::debug(const char *txt, ...)
{
    va_list ap;
    fprintf(logfile, "%sDEBUG: ", indentString);
    va_start(ap, txt);
    vfprintf(logfile, txt, ap);
#if !defined _WIN32
    printf("%sDEBUG: ", indentString);
    vprintf(txt, ap);
    fflush(stdout);
#endif
    va_end(ap);
    fflush(logfile);
}

void Logger::warning(const char *txt, ...)
{
    va_list ap;
    fprintf(logfile, "%sWARNING: ", indentString);
    va_start(ap, txt);
    vfprintf(logfile, txt, ap);
#if !defined _WIN32
    if (threshold > 0) {
        printf("%sWARNING: ", indentString);
        vprintf(txt,ap);
        fflush(stdout);
    }
#endif
    va_end(ap);
    fflush(logfile);
}

void Logger::note(const char *txt, ...)
{
    va_list ap;
    fprintf(logfile, "%s", indentString);
    va_start(ap, txt);
    vfprintf(logfile, txt, ap);
#if !defined _WIN32
    if (threshold > 1) {
        printf("%s", indentString);
        vprintf(txt,ap);
        fflush(stdout);
    }
#endif
    va_end(ap);
    fflush(logfile);
}

void Logger::gameMsg(const char *txt, ...)
{
    va_list ap;
    char msgstring[1024];
    fprintf(logfile, "%sGame: ", indentString);
    va_start(ap, txt);
    vfprintf(logfile, txt, ap);
    va_end(ap);
    fprintf(logfile, "\n");
    fflush(logfile);

    if(rendergamemsg && (msgpool != NULL)) {
        va_start(ap, txt);
        vsprintf(msgstring, txt, ap);
        va_end(ap);

        msgpool->postMessage(msgstring);
    } else {
#if !defined _WIN32
        printf("%s",indentString);
        vprintf(txt,ap);
        printf("\n");
        fflush(stdout);
#endif
    }
}


