#include "SDL_endian.h"
#include "externalvfs.h"
#include "common.h"
#include "memtrack.h"

using namespace std;

ExternalFiles::ExternalFiles(char *defpath)
{
    this->defpath = cppstrdup(defpath);
}

ExternalFiles::~ExternalFiles()
{
    unloadArchives();
    delete[] defpath;
}
    
bool ExternalFiles::loadArchive(const char *fname)
{
    if( isRelativePath(fname) ){
	char *pth;
	pth = new char[strlen(defpath)+strlen(fname)+2];
	sprintf(pth, "%s/%s", defpath, fname);
	path.push_back(pth);
	return true;
    }
    path.push_back(cppstrdup(fname));
    return true;
}

void ExternalFiles::unloadArchives()
{
    Uint32 i;

    for( i = 0; i < path.size(); i++ ) {
        delete[] path[i];
    }
    path.resize(0);
}


Uint32 ExternalFiles::getFile(const char *fname)
{
    openFileExt newFile;
    FILE *f;
    Uint32 i;
    char filename[1024];
    Uint32 size, fnum;

    for( i = 0; i < path.size(); i++ ) {
        strcpy(filename, path[i]);
        strcat(filename, fname);
        f = fopen(filename, "rb");
        if( f != NULL ) {
            fseek(f, 0, SEEK_END);
            size = ftell(f);
            fseek(f, 0, SEEK_SET);
            newFile.file = f;
            newFile.size = size;

            newFile.path = cppstrdup(filename);

            fnum = (Uint32)f;
            openFiles[fnum] = newFile;
            return fnum;
        }
    }

    return (Uint32)-1;
}

void ExternalFiles::releaseFile(Uint32 file)
{
    fclose(openFiles[file].file);
    delete[] openFiles[file].path;
    openFiles.erase(file);
}

Uint32 ExternalFiles::readByte(Uint32 file, Uint8 *databuf, Uint32 numBytes)
{
    return fread(databuf, 1, numBytes, openFiles[file].file);
}

Uint32 ExternalFiles::readWord(Uint32 file, Uint16 *databuf, Uint32 numWords)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 2, numWords, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap16(databuf[i]);
    }
#endif

    return numRead;
}

Uint32 ExternalFiles::readThree(Uint32 file, Uint32 *databuf, Uint32 numThrees)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 3, numThrees, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap32(databuf[i]);
        databuf[i]<<=8;
    }
#endif

    return numRead;
}

Uint32 ExternalFiles::readDWord(Uint32 file, Uint32 *databuf, Uint32 numDWords)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 4, numDWords, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap32(databuf[i]);
    }
#endif

    return numRead;
}

char *ExternalFiles::readLine(Uint32 file, char *databuf, Uint32 buflen)
{
    return fgets(databuf, buflen, openFiles[file].file);
}

void ExternalFiles::seekSet(Uint32 file, Uint32 pos)
{
    fseek(openFiles[file].file, pos, SEEK_SET);
}

void ExternalFiles::seekCur(Uint32 file, Sint32 pos)
{
    fseek(openFiles[file].file, pos, SEEK_CUR);
}

Uint32 ExternalFiles::getPos(Uint32 file)
{
    return ftell(openFiles[file].file);
}

Uint32 ExternalFiles::getSize(Uint32 file)
{
    return openFiles[file].size;
}

char *ExternalFiles::getPath(Uint32 file)
{
    return openFiles[file].path;
}


