#ifdef _MSC_VER
# pragma warning (disable: 4503)
# pragma warning (disable: 4786)
#endif
#include "SDL_endian.h"
#include "vfs.h"
#include "inifile.h"
#include <vector>
#include <map>
#include <string.h>
#include "logger.h"
#include "archive.h"
#include "externalvfs.h"
#include "vfsplugman.h"
#include "common.h"
#include "memtrack.h"


using namespace std;

Archive *externals;
VFSPlugMan *plugins;

void VFS_Init(char *binpath)
{
    Uint32 i;
    //int required;
    INIFile *filesini;
    char *defpath;
    char *game, *fname;
    char tempstr[1024];
    Uint32 keynum;
    //INIKey key;

    logger->note("Assuming FreeCNC is installed in \"%s\"\n", binpath);

    externals = new ExternalFiles(binpath);

    externals->loadArchive("conf/");
#if defined _WIN32
#elif defined(macintosh)
#else
    externals->loadArchive("/etc/freecnc/");
#endif

    try {
        filesini = new INIFile("files.ini");
    } catch(char *fname) {
        logger->error("Unable to locate \"%s\"\n", fname);
        exit(1);
    }
    i = 1;
    do {
        sprintf(tempstr, "path%d", i++);
        defpath = filesini->readString("general", tempstr);
        if( defpath != NULL ) {
            if( defpath[strlen(defpath)-1] != '/' &&
                    defpath[strlen(defpath)-1] != '\\' ) {
                strcpy(tempstr, defpath);
                strcat(tempstr, "/");
                delete[] defpath;
                defpath = cppstrdup(tempstr);
            }
            externals->loadArchive(defpath);
            delete[] defpath;
        }
    } while(defpath != NULL);
    /* load all the mixes here */
    for( i = 1;;i++ ) {
        plugins = new VFSPlugMan(binpath);
        sprintf(tempstr, "game%d", i);
        game = filesini->readString("general", tempstr);
        if( game == NULL ) {
            break;
        }
        //filesini->seekSection(game);
	try{
	    for( keynum = 1; ;keynum++ ){
		sprintf(tempstr, "required%d", keynum);
		fname = filesini->readString(game, tempstr);
		if( fname == NULL ){
		    delete[] fname;
		    break;
		}
		if( !plugins->loadArchive(fname) ){
                    logger->error("Missing required file \"%s\"\n", fname);
		    delete[] fname;
		    throw 0;
		}
		delete[] fname;
	    }
	}catch(int){
	    delete[] game;
	    delete plugins;
	    continue;
	}
	for( keynum = 1; ;keynum++ ){
	    sprintf(tempstr, "optional%d", keynum);
	    fname = filesini->readString(game, tempstr);
	    if( fname == NULL ){
		delete[] fname;
		break;
	    }
	    plugins->loadArchive(fname);
	    delete[] fname;
	}
	delete[] game;
	delete filesini;
	return;
/*
	try{
        for( keynum = 0;;keynum++ ) {
	    key = filesini->readKeyValue(game, keynum);
            //game = filesini->getLine(false);
            //if( game != NULL ) {
            sscanf(key->second.c_str(), "%d", &required );
            if( !plugins->loadArchive(key->first.c_str()) ) {
                if( required ) {
                    logger->error("Missing required file \"%s\"\n", key->first.c_str());
                    delete plugins;
                    delete[] game;
		    game = NULL;
                    //break;
		    throw 0;
                }
            }
        }
	}catch(int){
	}
        if( game != NULL ) {
	    delete[] game;
            delete filesini;
            return;
        }*/
	
    }

    delete filesini;
    logger->error("Unable to find mixes for any of the specified games!\n");
    exit(1);
}

void VFS_Destroy()
{
    delete plugins;
    delete externals;
}

VFile *VFS_Open(const char *fname)
{
    Uint32 fnum;
    fnum = externals->getFile(fname);
    if( fnum != (Uint32)-1 ) {
        return new VFile(fnum, externals);
    }
    return plugins->openFile(fname);
}

void VFS_Close(VFile *file)
{
    delete file;
}

